// ElemAry_UNS.cpp: CElemAry_Uns NX̃Cve[V
//
//			Copy Rights (c) Nobuyuki Umetani 2007
//          e-mail : umetani@sml.k.u-tokyo.ac.jp 
//			Last Updated 21/Feb/2007		
//////////////////////////////////////////////////////////////////////

#pragma warning ( disable : 4786 )
#pragma warning ( disable : 4996 )
#define for if(0);else for

#include <fstream>
#include <iostream>
#include <vector>
#include <string>
#include <assert.h>
#include <algorithm>

#include "ElemAry_UNS.h"

using namespace field;


////////////////
// Point
const unsigned int nlpoelPoint = 1;


////////////////
// LINE
const unsigned int nlpoelLine = 2;

const unsigned int nfaelLine = 2;
const unsigned int mxlpofaLine = 1;
const unsigned int nlpofaLine[nfaelLine] = {
	1, 1
};
const unsigned int lpofaLine[nfaelLine*mxlpofaLine] = {
	0, 
	1
};

////////////////
// Triangle
const unsigned int nlpoelTri = 3;

const unsigned int nfaelTri = 3;
const unsigned int mxlpofaTri = 2;
const unsigned int nlpofaTri[nfaelTri] = {
	2, 2, 2
};
const unsigned int lpofaTri[nfaelTri*mxlpofaTri] = {
	1, 2,
	2, 0,
	0, 1 
};


////////////////
// Quadric
const unsigned int nlpoelQuad = 4;

const unsigned int nfaelQuad = 4;
const unsigned int mxlpofaQuad = 2;
const unsigned int nlpofaQuad[nfaelQuad] = {
	2, 2, 2, 2
};
const unsigned int lpofaQuad[nfaelQuad*mxlpofaQuad] = {
	0, 1, 
	1, 2, 
	2, 3, 
	3, 0
};


////////////////
// Tetrahedra
const unsigned int nlpoelTet = 4;
const unsigned int nfaelTet = 4;
const unsigned int mxlpofaTet = 3;
const unsigned int nlpofaTet[nfaelTet] = {
	3, 3, 3, 3
};
const unsigned int lpofaTet[nfaelTet*mxlpofaTet] = {
	1, 2, 3,
	0, 3, 2,
	0, 1, 3,
	0, 2, 1
};


////////////////
// Hexahedra
const unsigned int nlpoelHex = 8;

const unsigned int nfaelHex = 6;
const unsigned int mxlpofaHex = 4;
const unsigned int nlpofaHex[nfaelHex] = {
	4, 4, 4, 4, 4, 4
};
const unsigned int lpofaHex[nfaelHex*mxlpofaHex] = {
	0, 3, 2, 1,
	0, 1, 5, 4,
	1, 2, 6, 5, 
	2, 3, 7, 6,
	3, 0, 4, 7,
	4, 5, 6, 7,
};

const unsigned int mxlp2lpedHex = 3;
const unsigned int nlp2lpedHex[nlpoelHex] = {
	3, 3, 3, 3, 3, 3, 3, 3
};
const unsigned int lp2lpedHex[nlpoelHex*mxlp2lpedHex] = {
	1, 3, 4,
	0, 2, 5,
	1, 3, 6,
	0, 2, 7, 
	0, 5, 7,
	1, 4, 6, 
	2, 5, 7, 
	3, 4, 6
};

struct SElemInfo{
	unsigned int nlpoel;

	unsigned int nfael;
	unsigned int mxlpofa;
	const unsigned int* nlpofa;
	const unsigned int* lpofa;

	unsigned int  mxlp2lped;
	const unsigned int* nlp2lped;
	const unsigned int* lp2lped;
};

const SElemInfo ElemInfoAry[6] = {
	{	// 0
		nlpoelPoint
	},
	{	// 1
		nlpoelLine,
		nfaelLine,  mxlpofaLine,  nlpofaLine,  lpofaLine
	},
	{	// 2
		nlpoelTri,
		nfaelTri,  mxlpofaTri,  nlpofaTri,  lpofaTri
	},
	{	// 3
		nlpoelQuad,
		nfaelQuad, mxlpofaQuad, nlpofaQuad, lpofaQuad
	},
	{	// 4
		nlpoelTet,
		nfaelTet,  mxlpofaTet,  nlpofaTet,  lpofaTet
	},
	{	// 5
		nlpoelHex,
		nfaelHex,  mxlpofaHex,  nlpofaHex,  lpofaHex,
		mxlp2lpedHex, nlp2lpedHex, lp2lpedHex
	}
};



bool CElemAry_UNS::Init(std::string& str_elem_type, const unsigned int& nelem, const unsigned int& npoel, const unsigned int* lnods ){
	this->nelem = nelem;
	this->npoel = npoel;
	if( this->m_pLnods != 0 ){ delete[] this->m_pLnods; }
	this->m_pLnods = new unsigned int [nelem*npoel];
	for(unsigned int i=0;i<nelem*npoel;i++){
		this->m_pLnods[i] = lnods[i];
	}
	end_ipoin = 0;
	for(unsigned int i=0;i<nelem*npoel;i++){
		end_ipoin = ( m_pLnods[i] > end_ipoin ) ? m_pLnods[i] : end_ipoin;
	}
	end_ipoin = end_ipoin+1;
	if(      str_elem_type == "LINE" ){ elem_type = LINE; }
	else if( str_elem_type == "TRI"  ){ elem_type = TRI;  }
	else if( str_elem_type == "QUAD" ){ elem_type = QUAD; }
	else if( str_elem_type == "TET"  ){ elem_type = TET;  }
	else if( str_elem_type == "HEX"  ){ elem_type = HEX;  }
	else if( str_elem_type == "POINT"){ elem_type = POINT;}
	else{
		std::cout << "Error!-->Not Implimented (Elem Type:" << str_elem_type << ")" << std::endl;
		assert(0);
		abort();
	}
	return true;
}


CElemAry* CElemAry_UNS::MakeBoundElemAry() const
{
	const unsigned int nlpoel = ElemInfoAry[ elem_type ].nlpoel;
	const unsigned int nfael = ElemInfoAry[ elem_type ].nfael;
	const unsigned int* lpofa = ElemInfoAry[ elem_type ].lpofa;
	const unsigned int* nlpofa = ElemInfoAry[ elem_type ].nlpofa;
	const unsigned int mxlpofa = ElemInfoAry[ elem_type ].mxlpofa;

	unsigned int* elsuel = new unsigned int [nelem*nfael];
	this->MakeElemSurElem(elsuel);

	std::string str_tmp_elem_type;
	if( elem_type == TRI || elem_type == QUAD ){
		str_tmp_elem_type = "LINE";
	}
	else if( elem_type == TET ){
		str_tmp_elem_type = "TRI";
	}
	else if( elem_type == HEX ){
		str_tmp_elem_type = "QUAD";
	}
	const unsigned int tmp_nlpoel = mxlpofa;

	unsigned int tmp_nelem;
	unsigned int* tmp_lnods;
	{
		unsigned int icoun0 = 0; 
		for(unsigned int ielem=0;ielem<nelem;ielem++){
			for(unsigned int ifael=0;ifael<nfael;ifael++){
				if( elsuel[ielem*nfael+ifael] == -1 ) icoun0++;
			}
		}
		tmp_nelem = icoun0;
		tmp_lnods = new unsigned int [tmp_nelem*tmp_nlpoel];
		icoun0 = 0;
		for(unsigned int ielem=0;ielem<nelem;ielem++){
			for(unsigned int ifael=0;ifael<nfael;ifael++){
				if( elsuel[ielem*nfael+ifael] == -1 ){
					const unsigned int nilpofa = nlpofa[ifael];
					for(unsigned int ilpofa=0;ilpofa<nilpofa;ilpofa++){
						tmp_lnods[icoun0*mxlpofa+ilpofa] 
							= m_pLnods[ielem*nlpoel+ lpofa[ifael*mxlpofa+ilpofa] ];
					}
					icoun0++;
				}
			}
		}
	}

	CElemAry_UNS* tmp_ea = new CElemAry_UNS;
	tmp_ea->Init(str_tmp_elem_type,tmp_nelem,tmp_nlpoel,tmp_lnods);

	delete tmp_lnods;

	return tmp_ea;
}

bool CElemAry_UNS::MakeEdge(unsigned int& nedge, unsigned int*& edge_ary) const{
	if( elem_type == TET ){
		const unsigned int npoin = end_ipoin;

		unsigned int* elsup_ind = new unsigned int [npoin+1];
		unsigned int* elsup = 0;
		unsigned int nelsup;
		this->MakePointSurElem(npoin,  elsup_ind,nelsup,elsup);

		unsigned int* psup_ind = new unsigned int [npoin+1];
		unsigned int npsup;
		unsigned int* psup;
		this->MakePointSurPoint(npoin,elsup_ind,nelsup,elsup,  psup_ind,npsup,psup );

		delete[] elsup_ind;	delete[] elsup;

		unsigned int icoun0 = 0;
		for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
			for(unsigned ipsup=psup_ind[ipoin];ipsup<psup_ind[ipoin+1];ipsup++){
				unsigned int jpoin0 = psup[ipsup];
				if( ipoin < jpoin0 ) icoun0++;
			}
		}

		nedge = icoun0;
		edge_ary = new unsigned int [nedge*2];

		icoun0 = 0;
		for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
			for(unsigned ipsup=psup_ind[ipoin];ipsup<psup_ind[ipoin+1];ipsup++){
				unsigned int jpoin0 = psup[ipsup];
				if( ipoin < jpoin0 ){
					edge_ary[icoun0*2  ] = ipoin;
					edge_ary[icoun0*2+1] = jpoin0;
					icoun0++;
				}
			}
		}

		delete[] psup_ind;	delete[] psup;

		return true;
	}
	else if( elem_type == TRI ){
		const SElemInfo& elem_info = ElemInfoAry[ elem_type ];
		const unsigned int nfael = elem_info.nfael;
		assert( nfael == 3 );
		unsigned int* elsuel = new unsigned int [nelem*nfael];
		this->MakeElemSurElem(elsuel);

		unsigned int icoun0 = 0;
		for(int ielem=0;ielem<nelem;ielem++){
			if( m_pLnods[ielem*npoel  ] > m_pLnods[ielem*npoel+1] || elsuel[ielem*nfael+2] == -1 ){ icoun0++; }
			if( m_pLnods[ielem*npoel+1] > m_pLnods[ielem*npoel+2] || elsuel[ielem*nfael  ] == -1 ){ icoun0++; }
			if( m_pLnods[ielem*npoel+2] > m_pLnods[ielem*npoel  ] || elsuel[ielem*nfael+1] == -1 ){ icoun0++; }
		}
		nedge = icoun0;
		edge_ary = new unsigned int [nedge*2];

		icoun0 = 0;
		for(int ielem=0;ielem<nelem;ielem++){
			if( m_pLnods[ielem*npoel  ] > m_pLnods[ielem*npoel+1] || elsuel[ielem*nfael+2] == -1 ){
				edge_ary[icoun0*2  ] = m_pLnods[ielem*npoel  ]; 
				edge_ary[icoun0*2+1] = m_pLnods[ielem*npoel+1];
				icoun0++;
			}
			if( m_pLnods[ielem*npoel+1] > m_pLnods[ielem*npoel+2] || elsuel[ielem*nfael  ] == -1 ){
				edge_ary[icoun0*2  ] = m_pLnods[ielem*npoel+1];
				edge_ary[icoun0*2+1] = m_pLnods[ielem*npoel+2];
				icoun0++;
			}
			if( m_pLnods[ielem*npoel+2] > m_pLnods[ielem*npoel  ] || elsuel[ielem*nfael+1] == -1 ){
				edge_ary[icoun0*2  ] = m_pLnods[ielem*npoel+2];
				edge_ary[icoun0*2+1] = m_pLnods[ielem*npoel  ];
				icoun0++;
			}
		}
		assert( nedge == icoun0 );

		delete[] elsuel;
		return true;
	}
	else if( elem_type == QUAD ){
		const SElemInfo& elem_info = ElemInfoAry[ elem_type ];
		const unsigned int nfael = elem_info.nfael;
		assert( nfael == 4 );
		unsigned int* elsuel = new unsigned int [nelem*nfael];
		this->MakeElemSurElem(elsuel);

		unsigned int icoun0 = 0;

		for(int ielem=0;ielem<nelem;ielem++){
			if( m_pLnods[ielem*npoel  ] > m_pLnods[ielem*npoel+1] || elsuel[ielem*nfael  ] == -1 ){ icoun0++; }
			if( m_pLnods[ielem*npoel+1] > m_pLnods[ielem*npoel+2] || elsuel[ielem*nfael+1] == -1 ){ icoun0++; }
			if( m_pLnods[ielem*npoel+2] > m_pLnods[ielem*npoel+3] || elsuel[ielem*nfael+2] == -1 ){ icoun0++; }
			if( m_pLnods[ielem*npoel+3] > m_pLnods[ielem*npoel  ] || elsuel[ielem*nfael+3] == -1 ){ icoun0++; }
		}
		nedge = icoun0;
		edge_ary = new unsigned int [nedge*2];

		icoun0 = 0;
		for(int ielem=0;ielem<nelem;ielem++){
			if( m_pLnods[ielem*npoel  ] > m_pLnods[ielem*npoel+1] || elsuel[ielem*nfael  ] == -1 ){
				edge_ary[icoun0*2  ] = m_pLnods[ielem*npoel  ];
				edge_ary[icoun0*2+1] = m_pLnods[ielem*npoel+1];
				icoun0++;
			}
			if( m_pLnods[ielem*npoel+1] > m_pLnods[ielem*npoel+2] || elsuel[ielem*nfael+1] == -1 ){
				edge_ary[icoun0*2  ] = m_pLnods[ielem*npoel+1];
				edge_ary[icoun0*2+1] = m_pLnods[ielem*npoel+2];
				icoun0++;
			}
			if( m_pLnods[ielem*npoel+2] > m_pLnods[ielem*npoel+3] || elsuel[ielem*nfael+2] == -1 ){
				edge_ary[icoun0*2  ] = m_pLnods[ielem*npoel+2];
				edge_ary[icoun0*2+1] = m_pLnods[ielem*npoel+3];
				icoun0++;
			}
			if( m_pLnods[ielem*npoel+3] > m_pLnods[ielem*npoel  ] || elsuel[ielem*nfael+3] == -1 ){
				edge_ary[icoun0*2  ] = m_pLnods[ielem*npoel+3];
				edge_ary[icoun0*2+1] = m_pLnods[ielem*npoel  ];
				icoun0++;
			}
		}
		assert( nedge == icoun0 );

		delete[] elsuel;
		return true;
	}
	else if( elem_type == HEX ){
		const unsigned int npoin = end_ipoin;

		unsigned int* elsup_ind = 0;
		unsigned int nelsup;
		unsigned int* elsup2 = 0;
		{
			elsup_ind = new unsigned int [npoin+1];
			for(unsigned int ipoin=0;ipoin<npoin+1;ipoin++){
				elsup_ind[ipoin] = 0;
			}
			for(unsigned int ielem=0;ielem<nelem;ielem++){
				for(unsigned int inoel=0;inoel<npoel;inoel++){
					const unsigned int ipoin0 = m_pLnods[ielem*npoel+inoel];
					assert( ipoin0 < npoin );
					elsup_ind[ipoin0+1]++;
				}
			}
			for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
				elsup_ind[ipoin+1] += elsup_ind[ipoin];
			}
			nelsup = elsup_ind[npoin];
			elsup2 = new unsigned int [nelsup*2];
			for(unsigned int ielem=0;ielem<nelem;ielem++){
				for(unsigned int inoel=0;inoel<npoel;inoel++){
					const unsigned int ipoin0 = m_pLnods[ielem*npoel+inoel];
					assert( ipoin0 < npoin );
					const unsigned int ipsup0 = elsup_ind[ipoin0];
					assert( ipsup0 < nelsup );
					elsup2[ipsup0*2  ] = ielem;
					elsup2[ipsup0*2+1] = inoel;
					elsup_ind[ipoin0]++;
				}
			}
			for(unsigned int ipoin=npoin;ipoin>0;ipoin--){
				elsup_ind[ipoin] = elsup_ind[ipoin-1];
			}
			elsup_ind[0] = 0;
		}

		unsigned int* psup_ind = new unsigned int [npoin+1];
		unsigned int npsup;
		unsigned int* psup = 0;
		{
			const unsigned int mxlp2lped = ElemInfoAry[ elem_type ].mxlp2lped;
			const unsigned int* nlp2lped = ElemInfoAry[ elem_type ].nlp2lped;
			const unsigned int* lp2lped = ElemInfoAry[ elem_type ].lp2lped;

			int* tmp_poin = new int [npoin];
			for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
				tmp_poin[ipoin] = -1;
			}
			int icoun0 = 0;
			for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
				tmp_poin[ipoin] = ipoin;
				for(unsigned int ielsup=elsup_ind[ipoin];ielsup<elsup_ind[ipoin+1];ielsup++){
					const unsigned int ielem0 = elsup2[ielsup*2  ];
					const unsigned int ipoel0 = elsup2[ielsup*2+1];
					const unsigned int nilp2lped = nlp2lped[ipoel0];
					for(unsigned int ilp2lped=0;ilp2lped<nilp2lped;ilp2lped++){
						const unsigned int jpoel0 = lp2lped[ipoel0*mxlp2lped+ilp2lped];
						const unsigned int jpoin0 = m_pLnods[ielem0*npoel+jpoel0];
						if( tmp_poin[jpoin0] != ipoin ){
							tmp_poin[jpoin0] = ipoin;
							icoun0++;
						}
					}
				}
			}

			npsup = icoun0;
			psup = new unsigned int [npsup];

			icoun0 = 0;
			psup_ind[0] = 0;
			for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
				tmp_poin[ipoin] = ipoin;
				for(unsigned int ielsup=elsup_ind[ipoin];ielsup<elsup_ind[ipoin+1];ielsup++){
					const unsigned int ielem0 = elsup2[ielsup*2  ];
					const unsigned int ipoel0 = elsup2[ielsup*2+1];
					const unsigned int nilp2lped = nlp2lped[ipoel0];
					for(unsigned int ilp2lped=0;ilp2lped<nilp2lped;ilp2lped++){
						const unsigned int jpoel0 = lp2lped[ipoel0*mxlp2lped+ilp2lped];
						const unsigned int jpoin0 = m_pLnods[ielem0*npoel+jpoel0];
						if( tmp_poin[jpoin0] != ipoin ){
							tmp_poin[jpoin0] = ipoin;
							psup[icoun0] = jpoin0;
							icoun0++;
						}
					}
				}
				psup_ind[ipoin+1] = icoun0;
			}
			delete[] tmp_poin;
		}

		delete[] elsup_ind;	delete[] elsup2;

		unsigned int icoun0 = 0;
		for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
			for(unsigned ipsup=psup_ind[ipoin];ipsup<psup_ind[ipoin+1];ipsup++){
				unsigned int jpoin0 = psup[ipsup];
				if( ipoin < jpoin0 ) icoun0++;
			}
		}

		nedge = icoun0;
		edge_ary = new unsigned int [nedge*2];

		icoun0 = 0;
		for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
			for(unsigned ipsup=psup_ind[ipoin];ipsup<psup_ind[ipoin+1];ipsup++){
				unsigned int jpoin0 = psup[ipsup];
				if( ipoin < jpoin0 ){
					edge_ary[icoun0*2  ] = ipoin;
					edge_ary[icoun0*2+1] = jpoin0;
					icoun0++;
				}
			}
		}

		delete[] psup_ind;	delete[] psup;
	}
	else{
		std::cout << "Error!-->Not Implimented" << std::endl;
		std::cout << "Elem Type : " << elem_type << std::endl;
		assert(0);
		abort();
			nedge = 0;
	}
	return true;
}

bool CElemAry_UNS::MakeElemSurElem(unsigned int* elsuel) const
{
	const unsigned int npoin = end_ipoin;
	unsigned int* elsup_ind = new unsigned int [npoin+1];
	unsigned int* elsup;
	unsigned int nelsup;
	this->MakePointSurElem(npoin, elsup_ind,nelsup,elsup);
	this->MakeElemSurElem(npoin,elsup_ind,nelsup,elsup,elsuel);
	delete[] elsup_ind;
	delete[] elsup;
	return true;
}

bool CElemAry_UNS::MakeElemSurElem(const unsigned int npoin, 
					 const unsigned int* elsup_ind, const unsigned int nelsup, const unsigned int* elsup,
					 unsigned int* elsuel ) const{

	assert( npoin >= end_ipoin );

	SElemInfo elem_info = ElemInfoAry[elem_type];

	const unsigned int nfael = elem_info.nfael;
	const unsigned int mxlpofa = elem_info.mxlpofa;
	const unsigned int* nlpofa = elem_info.nlpofa;
	const unsigned int* lpofa = elem_info.lpofa;

	unsigned int* tmp_poin = new unsigned int [npoin];
	for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
		tmp_poin[ipoin] = 0;
	}
	unsigned int inpofa[4];

	for(unsigned int ielem=0;ielem<nelem;ielem++){
		for(unsigned int ifael=0;ifael<nfael;ifael++){
			const unsigned int nilpofa = nlpofa[ifael];
			for(unsigned int ipofa=0;ipofa<nilpofa;ipofa++){
				inpofa[ipofa] = m_pLnods[ielem*npoel+ lpofa[ifael*mxlpofa+ipofa] ];
				tmp_poin[ inpofa[ipofa] ] = 1;
			}
			const unsigned int ipoin0= inpofa[0];
			bool iflg = false;
			for(unsigned int ielsup=elsup_ind[ipoin0];ielsup<elsup_ind[ipoin0+1];ielsup++){
				const unsigned int jelem0 = elsup[ielsup];
				if( jelem0 == ielem ) continue;
				for(unsigned int jfael=0;jfael<nfael;jfael++){
					iflg = true;
					const unsigned int njlpofa = nlpofa[jfael];
					for(unsigned int jpofa=0;jpofa<njlpofa;jpofa++){
						const unsigned int jpoin0 =  m_pLnods[jelem0*npoel + lpofa[jfael*mxlpofa+jpofa] ];
						if( tmp_poin[ jpoin0 ] == 0 ){
							iflg = false;
							break;
						}
					}
					if( iflg ){
						elsuel[ielem*nfael+ifael] = jelem0;
						break;
					}
				}
				if( iflg ) break;
			}
			if( !iflg ){ elsuel[ielem*nfael+ifael] = -1; }
			for(unsigned int ipofa=0;ipofa<nilpofa;ipofa++){
				tmp_poin[ inpofa[ipofa] ] = 0;
			}
		}
	}

	delete[] tmp_poin;
	return true;
}

bool CElemAry_UNS::MakePointSurElem(
	const unsigned int npoin,
	unsigned int* const elsup_ind,
	unsigned int& nelsup,
	unsigned int*& elsup ) const {

	assert( npoin >= end_ipoin );

	for(unsigned int ipoin=0;ipoin<npoin+1;ipoin++){
		elsup_ind[ipoin] = 0;
	}
	for(unsigned int ielem=0;ielem<nelem;ielem++){
		for(unsigned int inoel=0;inoel<npoel;inoel++){
			const unsigned int ipoin0 = m_pLnods[ielem*npoel+inoel];
			assert( ipoin0 < npoin );
			elsup_ind[ipoin0+1]++;
		}
	}
	for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
		elsup_ind[ipoin+1] += elsup_ind[ipoin];
	}
	nelsup = elsup_ind[npoin];
	elsup = new unsigned int [nelsup];
	for(unsigned int ielem=0;ielem<nelem;ielem++){
		for(unsigned int inoel=0;inoel<npoel;inoel++){
			const unsigned int ipoin0 = m_pLnods[ielem*npoel+inoel];
			assert( ipoin0 < npoin );
			const unsigned int ipsup0 = elsup_ind[ipoin0];
			assert( ipsup0 < nelsup );
			elsup[ipsup0] = ielem;
			elsup_ind[ipoin0]++;
		}
	}
	for(unsigned int ipoin=npoin;ipoin>0;ipoin--){
		elsup_ind[ipoin] = elsup_ind[ipoin-1];
	}
	elsup_ind[0] = 0;
/*
	for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
		std::cout << ipoin << "--> ";
		for(unsigned int ielsup=elsup_ind[ipoin];ielsup<elsup_ind[ipoin+1];ielsup++){
			std::cout << elsup[ielsup] << " ";
		}
		std::cout << std::endl;
	}
*/
	return true;
}


bool CElemAry_UNS::MakePointSurPoint(const unsigned int& npoin, 
								 unsigned int* const psup_ind, unsigned int& npsup, unsigned int*& psup ) const {
	assert( npoin >= end_ipoin );
	unsigned int* elsup_ind = new unsigned int [npoin+1];
	unsigned int* elsup;
	unsigned int nelsup;
	this->MakePointSurElem(npoin, elsup_ind,nelsup,elsup);
	this->MakePointSurPoint(npoin,elsup_ind,nelsup,elsup,  psup_ind,npsup,psup);
	delete[] elsup_ind;
	delete[] elsup;
	return true;
}

bool CElemAry_UNS::MakePointSurPoint(const unsigned int npoin, 
								 const unsigned int* elsup_ind, const unsigned int nelsup, const unsigned int* elsup,
								 unsigned int* const psup_ind, unsigned int& npsup, unsigned int*& psup ) const {

	int* lpoin = new int [npoin];
	for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
		lpoin[ipoin] = -1;
	}
	int icoun0 = 0;
	for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
		lpoin[ipoin] = ipoin;
		for(unsigned int ielsup=elsup_ind[ipoin];ielsup<elsup_ind[ipoin+1];ielsup++){
			const unsigned int jelem0 = elsup[ielsup];
			for(int jlpoel=0;jlpoel<npoel;jlpoel++){
				const unsigned int jpoin0 = m_pLnods[jelem0*npoel+jlpoel];
				if( lpoin[jpoin0] != ipoin ){
					lpoin[jpoin0] = ipoin;
					icoun0++;
				}
			}
		}
	}

	npsup = icoun0;
	psup = new unsigned int [npsup];

	icoun0 = 0;
	psup_ind[0] = 0;
	for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
		lpoin[ipoin] = ipoin;
		for(unsigned int ielsup=elsup_ind[ipoin];ielsup<elsup_ind[ipoin+1];ielsup++){
			const unsigned int jelem0 = elsup[ielsup];
			for(int jlpoel=0;jlpoel<npoel;jlpoel++){
				const unsigned int jpoin0 = m_pLnods[jelem0*npoel+jlpoel];
				if( lpoin[jpoin0] != ipoin ){
					lpoin[jpoin0] = ipoin;
					psup[icoun0] = jpoin0;
					icoun0++;
				}
			}
		}
		psup_ind[ipoin+1] = icoun0;
	}
	delete[] lpoin;

	for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
		std::sort(&psup[ psup_ind[ipoin] ], &psup[psup_ind[ipoin+1] ]);
	}
/*
	for(unsigned int ipoin=0;ipoin<npoin;ipoin++){
		std::cout << ipoin << "--> ";
		for(unsigned int ipsup=psup_ind[ipoin];ipsup<psup_ind[ipoin+1];ipsup++){
			std::cout << psup[ipsup] << " ";
		}
		std::cout << std::endl;
	}
*/
	return true;
}

int CElemAry_UNS::WriteToFile(const std::string& file_name, long& offset, unsigned int id) const {
	
	FILE *fp;

	if( (fp = fopen(file_name.c_str(),"a"))== NULL ){
		std::cout << "Error!-->Cannot Open File" << std::endl;
		assert(0);
		return false;
	}

	fprintf(fp,"$$$$$$\n");
	fprintf(fp,"ELEM_ARY_UNS\n");

	fprintf(fp,"%d\n",id);

	unsigned int nnoel;
	if( this->elem_type == TRI ){		fprintf(fp,"TRI\n");  nnoel=3; }
	else if( this->elem_type == QUAD ){ fprintf(fp,"QUAD\n"); nnoel=4; }
	else if( this->elem_type == LINE ){ fprintf(fp,"LINE\n"); nnoel=2; }
	else if( this->elem_type == POINT){ fprintf(fp,"POINT\n"); nnoel=1;}
	else{ assert(0); }

/*
	std::vector<unsigned int> id_ary_ns = this->m_aSeg.GetAry_ObjID();
	fprintf(fp,"%d\n",id_ary_ns.size());
	for(unsigned int iid_ns=0;iid_ns<id_ary_ns.size();iid_ns++){
		unsigned int id_ns = id_ary_ns[iid_ns];
		const CNodeSeg& ns = m_aSeg.GetObj(id_ns);
		fprintf(fp,"%d %d %d %s\n", iid_ns+1, id_ns, ns.len, ns.name.c_str());
	}
*/

	fprintf(fp,"1\n");
	fprintf(fp,"1 1 %d CORNER\n",nnoel);

	fprintf(fp,"%d %d\n",this->Size(),nnoel);

	for(unsigned int ielem=0;ielem<this->Size();ielem++){
		fprintf(fp,"%d ",ielem+1);
		for(unsigned int inoel=0;inoel<nnoel;inoel++){
			fprintf(fp,"%d ",this->m_pLnods[ielem*nnoel+inoel]+1);
		}
		fprintf(fp,"\n");
	}

	fclose(fp);

	return 0;
}

int CElemAry_UNS::InitializeFromFile(const std::string& file_name, long& offset){

	FILE *fp;
	const unsigned int buff_size = 512;
	char stmp1[buff_size];

	if( (fp = fopen(file_name.c_str(),"r"))== NULL ){
		std::cout << "Error!-->Cannot Open File" << std::endl;
		assert(0);
		return false;
	}

	fseek(fp,offset,SEEK_SET);

	////////////////////////////////
	while( 1 ){
		if( fgets(stmp1,buff_size,fp) == NULL ){ return 1; }
		if( stmp1[0] == '#' ) continue;
		if( strspn(stmp1," \n") != strlen(stmp1) ) break;
	}
	assert( strncmp(stmp1,"$$$$$$$$",6)==0 );

	while( 1 ){
		if( fgets(stmp1,buff_size,fp) == NULL ){ return -1; }
		if( stmp1[0] == '#' ) continue;
		if( strspn(stmp1," \n") != strlen(stmp1) ) break;
	}
	assert( strncmp(stmp1,"ELEM_ARY_UNS",12)==0 );

	{
		while( fgets(stmp1,buff_size,fp) != NULL ){
			if( stmp1[0] == '#' ) continue;
			if( strspn(stmp1," \n") != strlen(stmp1) ) break;
		}
		int tmp_id;
		sscanf(stmp1,"%d",&tmp_id);
		assert( tmp_id > 0 );
		unsigned int id = tmp_id;
	}

	std::string str_elem_type;
	{
		while( fgets(stmp1,buff_size,fp) != NULL ){
			if( stmp1[0] == '#' ) continue;
			if( strspn(stmp1," \n") != strlen(stmp1) ) break;
		}
		const unsigned int len = strlen(stmp1);
		stmp1[len-1] = '\0';
		str_elem_type = stmp1;
	}

	// read nseg_ea
	unsigned int nseg;
	{
		while( fgets(stmp1,buff_size,fp) != NULL ){
			if( stmp1[0] == '#' ) continue;
			if( strspn(stmp1," \n") != strlen(stmp1) ) break;
		}
		int tmp_nseg;
		sscanf(stmp1,"%d",&tmp_nseg);
		assert( tmp_nseg > 0 );
		nseg = tmp_nseg;
	}

	std::vector<CElemSeg> es_ary;
	{
		while( fgets(stmp1,buff_size,fp) != NULL ){
			if( stmp1[0] == '#' ) continue;
			if( strspn(stmp1," \n") != strlen(stmp1) ) break;
		}
		int tmp_iseg, tmp_id, tmp_len, tmp_idna;
		char tmp_name[256];
		CElemSeg tmp_es;
		for(int iseg=0;;iseg++){
			sscanf(stmp1,"%d%d%d%d%s",&tmp_iseg, &tmp_id, &tmp_len, &tmp_idna, tmp_name);
			assert( tmp_iseg-1 == iseg );
			assert( tmp_id > 0 );
			assert( tmp_len > 0 );
			tmp_es.id = tmp_id;
			tmp_es.len = tmp_len;
			tmp_es.id_na = tmp_idna;
			es_ary.push_back(tmp_es);
			if( iseg >= nseg-1 ) break;
			fgets(stmp1,buff_size,fp);
		}
	}
		
	unsigned int nelem, nnoel;
	{
		while( fgets(stmp1,buff_size,fp) != NULL ){
			if( stmp1[0] == '#' ) continue;
			if( strspn(stmp1," \n") != strlen(stmp1) ) break;
		}
		int tmp_nelem, tmp_nnoel;
		sscanf(stmp1,"%d%d",&tmp_nelem, &tmp_nnoel);
		assert( tmp_nelem > 0 && tmp_nnoel > 0 );
		nelem = tmp_nelem;
		nnoel = tmp_nnoel;
	}

	unsigned int* lnods;
	{
		lnods = new unsigned int [nelem*nnoel];
		while( fgets(stmp1,buff_size,fp) != NULL ){
			if( stmp1[0] == '#' ) continue;
			if( strspn(stmp1," \n") != strlen(stmp1) ) break;
		}
		char *tp;
		int inode0;
		int tmp_ielem;
		for(unsigned int ielem=0;;ielem++){
			tp = strtok(stmp1," ");
			tmp_ielem = atoi(tp);
			assert( tmp_ielem-1 == ielem );
			for(unsigned int inoel=0;inoel<nnoel;inoel++){
				tp = strtok(NULL," ");
				inode0 = atoi(tp)-1; 
				assert( inode0 >= 0 );
				lnods[ielem*nnoel+inoel] = inode0;
			}
			if( ielem >= nelem-1 ) break;
			fgets(stmp1,buff_size,fp);
		}
	}

	offset = ftell(fp);
	fclose(fp);

	this->Init(str_elem_type,nelem,nnoel,lnods);

	delete[] lnods;

	return 0;
}