// Field.h : CField NX̃C^[tF[X
//
//			Copy Rights (c) Nobuyuki Umetani 2007
//          e-mail : umetani@sml.k.u-tokyo.ac.jp 
//			Last Updated 03/04/2007		
//////////////////////////////////////////////////////////////////////

#if !defined(FIELD_H)
#define FIELD_H

#pragma warning ( disable : 4996 )

#include <string>

#include "ElemArySet.h"
#include "NodeAry.h"
#include "../Mat/BCFlag_Blk.h"

namespace field{
	class CField{
	public:
		enum FIELD_TYPE{ SCALAR, VECTOR };
	public:
		CField(){
			m_id_es = 0;
			m_id_ea = 0;
			m_id_ns_co  = 0;
			m_id_ns_val = 0;
		}
		CField(
			const FIELD_TYPE& field_type, const std::string& name,
			unsigned int id_ea, 
			CNodeAry& na, CElemArySet& eas)
		{	
			m_field_type = field_type;
			m_name = name;
			m_id_ea = id_ea;
			m_id_ns_co = na.GetAry_SegID()[0];
			std::vector< std::pair<unsigned int,field::CNodeAry::CNodeSeg> > add_ns;
			if( field_type == SCALAR ){
				add_ns.push_back( std::make_pair(na.GetFreeSegID(), field::CNodeAry::CNodeSeg(1,name)) );
			}
			else if( field_type == VECTOR ){
				const CNodeAry::CNodeSeg& ns_co = na.GetSeg(m_id_ns_co);
				const unsigned int len = ns_co.len;
				add_ns.push_back( std::make_pair(na.GetFreeSegID(), field::CNodeAry::CNodeSeg(len,name)) );
			}
			else{
				std::cout << "Error!-->Undefined field_type " << std::endl;
				assert(0);
				abort();
			}
			std::vector<int> id_ary = na.AddSegment( add_ns, 0.0 );
			m_id_ns_val = id_ary[0];
		}
		CField(
			const CField& field, const std::string& name,
			unsigned int id_ea, 
			CNodeAry& na, CElemArySet& eas)
			: m_id_ea(id_ea)
		{			
			m_field_type = field.m_field_type;
			m_name = field.m_name;
			m_id_ns_val = field.m_id_ns_val;
			m_id_ns_co  = field.m_id_ns_co;
		}
		void SetValueRandom(CNodeAry& na, const CElemArySet& eas) const{
			const field::CNodeAry::CNodeSeg& ns_val = na.GetSeg(m_id_ns_val);
			const unsigned int nnode = na.Size();
			srand(0);
			for(int inode=0;inode<nnode;inode++){
				na.SetValueToNode(inode,m_id_ns_val,0,  rand()*2.0/(1.0+RAND_MAX)-1.0);
			}
		}
		void SetValue(double val, unsigned int idofns, CNodeAry& na, const CElemArySet& eas) const{
			const field::CNodeAry::CNodeSeg& ns_val = na.GetSeg(m_id_ns_val);
			const CElemAry& ea = eas.GetEA(m_id_ea);
			na.SetValueToNode(ea,m_id_ns_val,idofns,val);
		}
		void BoundaryCondition(unsigned int idofns, ls::CBCFlag& bc_flag, const CNodeAry& na, CElemArySet& eas) const{
			const CElemAry& ea = eas.GetEA(m_id_ea);
			bc_flag.SetBC(ea,idofns);
		}
		unsigned int GetID_ElemAry() const { return m_id_ea; }
		unsigned int GetID_NodeSeg_Coord() const { return m_id_ns_co;  }
		unsigned int GetID_NodeSeg_Value() const { return m_id_ns_val; }

		
		int WriteToFile(const std::string& file_name, long& offset, unsigned int id) const {
			
			FILE *fp;

			if( (fp = fopen(file_name.c_str(),"a"))== NULL ){
				std::cout << "Error!-->Cannot Open File" << std::endl;
				assert(0);
				return false;
			}

			fprintf(fp,"$$$$$$\n");
			fprintf(fp,"FIELD\n");

			fprintf(fp,"%d\n",id);
			fprintf(fp,"%s\n",m_name.c_str());

			if( m_field_type == SCALAR ){
				fprintf(fp,"SCALAR\n");
			}
			if( m_field_type == VECTOR ){
				fprintf(fp,"VECTOR\n");
			}
			else{
				assert(0);
			}

			fprintf(fp,"%d\n",m_id_ea);
			fprintf(fp,"%d\n",m_id_ns_co);
			fprintf(fp,"%d\n",m_id_ns_val);

			fclose(fp);
			return 0;
		}

		bool InitializeFromFile(const std::string& file_name, long& offset){

			FILE *fp;
			const unsigned int buff_size = 512;
			char stmp1[buff_size];

			if( (fp = fopen(file_name.c_str(),"r"))== NULL ){
				std::cout << "Error!-->Cannot Open File" << std::endl;
				assert(0);
				return false;
			}

			fseek(fp,offset,SEEK_SET);

			////////////////////////////////

			// read "$$$$$$"
			while( fgets(stmp1,buff_size,fp) != NULL ){
				if( stmp1[0] == '#' ) continue;
				if( strspn(stmp1," \n") != strlen(stmp1) ) break;
			}
			assert( strncmp(stmp1,"$$$$$$$$",6)==0 );

			// read "FIELD"
			while( fgets(stmp1,buff_size,fp) != NULL ){
				if( stmp1[0] == '#' ) continue;
				if( strspn(stmp1," \n") != strlen(stmp1) ) break;
			}
			assert( strncmp(stmp1,"FIELD",5)==0 );

			// read id_fd
			unsigned int id;
			{
				while( fgets(stmp1,buff_size,fp) != NULL ){
					if( stmp1[0] == '#' ) continue;
					if( strspn(stmp1," \n") != strlen(stmp1) ) break;
				}
				int tmp_id;
				sscanf(stmp1,"%d",&tmp_id);
				assert( tmp_id > 0 );
				id = tmp_id;
			}

			// read "NAME"
			{
				while( fgets(stmp1,buff_size,fp) != NULL ){
					if( stmp1[0] == '#' ) continue;
					if( strspn(stmp1," \n") != strlen(stmp1) ) break;
				}
				m_name = stmp1;
			}

			// read "FIELD"
			{
				while( fgets(stmp1,buff_size,fp) != NULL ){
					if( stmp1[0] == '#' ) continue;
					if( strspn(stmp1," \n") != strlen(stmp1) ) break;
				}
				std::cout << stmp1 << std::endl;
				if( strcmp(stmp1,"SCALAR\n") == 0 ){
					m_field_type = SCALAR;
				}
				else if( strcmp(stmp1,"VECTOR\n") == 0 ){
					m_field_type = VECTOR;
				}
				else{
					assert(0);
				}
			}

			// read id_ea
			{
				while( fgets(stmp1,buff_size,fp) != NULL ){
					if( stmp1[0] == '#' ) continue;
					if( strspn(stmp1," \n") != strlen(stmp1) ) break;
				}
				int tmp_id_ea;
				sscanf(stmp1,"%d",&tmp_id_ea);
				assert( tmp_id_ea > 0 );
				m_id_ea = tmp_id_ea;
			}

			// read id_ns_co
			{
				while( fgets(stmp1,buff_size,fp) != NULL ){
					if( stmp1[0] == '#' ) continue;
					if( strspn(stmp1," \n") != strlen(stmp1) ) break;
				}
				int tmp_id_ns_co;
				sscanf(stmp1,"%d",&tmp_id_ns_co);
				assert( tmp_id_ns_co > 0 );
				m_id_ns_co = tmp_id_ns_co;
			}

			// read id_ns_val
			{
				while( fgets(stmp1,buff_size,fp) != NULL ){
					if( stmp1[0] == '#' ) continue;
					if( strspn(stmp1," \n") != strlen(stmp1) ) break;
				}
				int tmp_id_ns_val;
				sscanf(stmp1,"%d",&tmp_id_ns_val);
				assert( tmp_id_ns_val > 0 );
				m_id_ns_val = tmp_id_ns_val;
			}

			fclose(fp);

			return 0;
		}
	private:
		unsigned int m_id_es;
		unsigned int m_id_ea;
		unsigned int m_id_ns_co;
		unsigned int m_id_ns_val;
		FIELD_TYPE m_field_type;
		std::string m_name;
	};


}

#endif
