// File.h : 
//
//			Copy Rights (c) Nobuyuki Umetani 2007
//          e-mail : umetani@sml.k.u-tokyo.ac.jp 
//			Last Updated 10.Jun.2007		
//////////////////////////////////////////////////////////////////////

#if !defined(FILE_H)
#define FILE_H

#include <fstream>

#include "NodeAry.h"
#include "ElemAry.h"
#include "ElemAry_UNS.h"
#include "ElemArySet.h"
#include "Field.h"

#include "../Mat/Vector_Blk.h"

void DataOutInp(std::string file_name,
	const unsigned int id_ns_coord, const unsigned  int id_ns_val,
	const field::CNodeAry& na, const field::CElemAry& ea)
{
	////////////////

	const int nnode = na.Size();
	const int nelem = ea.Size();

	std::ofstream fout;
	fout.setf( std::ios::scientific );
	fout.open( file_name.c_str() );
	if( fout.fail() ){
		std::cout << "t@CJ܂" << std::endl;
		abort();
	}

	//wb_ZOg
	{

		fout << 1 << "\n";
		fout << "data_geom\n";
		fout << "step1\n";
		fout << nnode << " " << nelem << "\n";
	}

	//WZOg
	{
		const field::CNodeAry::CNodeSeg& ns_co = na.GetSeg(id_ns_coord);
		const unsigned int ndim = ns_co.len;
		if( ndim == 2 ){
			for(int inode=0;inode<nnode;inode++){
				double val;
				fout.width(6);
				fout << inode+1 << " ";
				fout.width(13);
				na.GetValueFromNode(inode,id_ns_coord,0,val);
				fout << val << " ";
				fout.width(13);
				na.GetValueFromNode(inode,id_ns_coord,1,val);
				fout << val << " ";
				fout.width(13);
				fout << 0.0 << "\n";
			}
		}
		if( ndim == 3 ){
			for(int inode=0;inode<nnode;inode++){
				double val;
				fout.width(6);
				fout << inode+1 << " ";
				fout.width(13);
				na.GetValueFromNode(inode,id_ns_coord,0,val);
				fout << val << " ";
				fout.width(13);
				na.GetValueFromNode(inode,id_ns_coord,1,val);
				fout << val << " ";
				fout.width(13);
				na.GetValueFromNode(inode,id_ns_coord,2,val);
				fout << val << "\n";
			}
		}
	}

	//RlNeBreBAZOg
	{
		unsigned int nnoel = 0;
		std::string elem_shape_name;
		field::CElemAry::ELEM_TYPE et = ea.ElemType();
		switch(et){
		case field::CElemAry::TRI : 
				elem_shape_name = "tri";
				nnoel = 3;
				break;
		case field::CElemAry::QUAD :
				elem_shape_name = "quad";
				nnoel = 4;
				break;
		case field::CElemAry::HEX :
				elem_shape_name = "hex";
				nnoel = 8;
				break;
		default : 
				assert(0);
		}

		unsigned int noel[256];
		unsigned int tmp_nnoel;
		for(int ielem=0;ielem<nelem;ielem++){
			ea.GetLpoel(ielem,tmp_nnoel,noel);
			assert( tmp_nnoel == nnoel );
			fout.width(6);
			fout << ielem+1 << " " << 0 << " " << elem_shape_name << "  ";
			for(int inoel=0;inoel<nnoel;inoel++){
				fout.width(6);
				fout << noel[inoel]+1 << " ";
			}
			fout << "\n";
		}
	}

	{			
		const field::CNodeAry::CNodeSeg& ns_val = na.GetSeg(id_ns_val);
		//lwb_ZOg
		fout << ns_val.len << " " << 0 << "\n";
		fout << ns_val.len << " ";
		for(unsigned int idofns=0;idofns<ns_val.len;idofns++){
			fout << 1 << " ";
		}
		fout << "\n";
		for(unsigned int idofns=0;idofns<ns_val.len;idofns++){
			fout << ns_val.name << ", \n";
		}

		//lZOg
		for(unsigned int inode=0;inode<nnode;inode++){
			fout << inode+1 << " ";
			for(unsigned int idofns=0;idofns<ns_val.len;idofns++){
				double val;
				na.GetValueFromNode(inode,id_ns_val,idofns,val);
				fout << val << " ";
			}
			fout << "\n";
		}
	}
	fout.close();
}

bool ReadMesh(const std::string& file_name, field::CNodeAry& NA, field::CElemArySet& EAS){
	long offset = 0;

	NA.InitializeFromFile(file_name,offset);
	EAS.InitializeFromFile(file_name,offset);

	return true;
}

bool ReadMesh(const std::string& file_name, 
			  field::CNodeAry& NA, 
			  field::CElemArySet& EAS,
			  field::CField& field){
	long offset = 0;

	NA.InitializeFromFile(file_name,offset);
	EAS.InitializeFromFile(file_name,offset);
	field.InitializeFromFile(file_name,offset);

	return true;
}

bool WriteMesh(const std::string& file_name, 
			   const field::CNodeAry& NA, const field::CElemArySet& EAS,
			   const field::CField& field){
	long offset = 0;

	std::ofstream fin;
	fin.open(file_name.c_str(),std::ios::trunc);
	fin.close();

	NA.WriteToFile(file_name,offset);
	EAS.WriteToFile(file_name,offset);
	field.WriteToFile(file_name,offset,1);

	return true;
}

#endif
