// MatDia_Crs.cpp: CMatDia_BlkCrs NX̃Cve[V
//
//			Copy Rights (c) Nobuyuki Umetani 2007
//          e-mail : umetani@sml.k.u-tokyo.ac.jp 
//			Last Updated 02/21/2007				
//////////////////////////////////////////////////////////////////////

#pragma warning( disable : 4786 ) 
#define for if(0); else for

#include <iostream>
#include <cassert>
#include <math.h>
#include <vector>
#include <algorithm>

#include "MatDia_BlkCrs.h"
#include "Vector_Blk.h"
#include "BCFlag_Blk.h"
#include "Order_Blk.h"

#include "../field/ElemAry.h"

using namespace ls;

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

CMatDia_BlkCrs::CMatDia_BlkCrs(const unsigned int nblk_colrow, const unsigned int len_colrow)
:CMat_BlkCrs(nblk_colrow, len_colrow, nblk_colrow, len_colrow)
{
//	std::cout << "Construct : CMatDia_BlkCrs(nblk_colrow, len_colrow) " << nblk_colrow << " " << len_colrow << std::endl;
	m_valDia_Blk = new  double [NBlkMatCol() * LenBlkCol()*LenBlkRow()];
}

CMatDia_BlkCrs::~CMatDia_BlkCrs()
{
	if( m_valDia_Blk != 0 ){ delete[] m_valDia_Blk; m_valDia_Blk = 0; }
}

bool CMatDia_BlkCrs::DeletePattern(){
	CMat_BlkCrs::DeletePattern();
	return true;
}

bool CMatDia_BlkCrs::SetZero(){
	CMat_BlkCrs::SetZero();
	const unsigned int ni = NBlkMatCol()*LenBlkCol()*LenBlkCol();
	for(unsigned int i=0;i<ni;i++){
		m_valDia_Blk[i] = 0.0;
	}
	return true;
}

bool CMatDia_BlkCrs::Mearge(unsigned int nblkel_col, const unsigned int* blkel_col,
						 unsigned int nblkel_row, const unsigned int* blkel_row,
						 unsigned int blksize, const double* emat, unsigned int* tmp_buffer)
{
	assert( m_valCrs_Blk != 0 );
	assert( m_valDia_Blk != 0 );

	const unsigned int BlkSize = LenBlkCol()*LenBlkRow();
	const unsigned int BlkLen = LenBlkCol();

	assert( nblkel_col == nblkel_row );
	assert( blksize == BlkSize );

	for(unsigned int iblkel=0;iblkel<nblkel_col;iblkel++){
		const int iblk1 = blkel_col[iblkel];
		for(unsigned int jpsup=m_colInd_Blk[iblk1];jpsup<m_colInd_Blk[iblk1+1];jpsup++){
			assert( jpsup < m_ncrs_Blk );
			const unsigned int jblk1 = m_rowPtr_Blk[jpsup];
			tmp_buffer[jblk1] = jpsup;
		}
		for(unsigned int jblkel=0;jblkel<nblkel_row;jblkel++){
			if( iblkel == jblkel ){	// Marge Diagonal
				const double* pval_in = &emat[(iblkel*nblkel_row+iblkel)*BlkSize];
				double* pval_out = &m_valDia_Blk[iblk1*BlkSize];
				for(unsigned int idof=0;idof<BlkSize;idof++){ pval_out[idof] += pval_in[idof]; }
			}
			else{	// Marge Non-Diagonal
				const unsigned int jblk1 = blkel_row[jblkel];
				const unsigned int jpsup1 = tmp_buffer[jblk1];
				assert( jpsup1 < m_ncrs_Blk );
				assert( m_rowPtr_Blk[jpsup1] == jblk1 );
				const double* pval_in = &emat[(iblkel*nblkel_row+jblkel)*BlkSize];
				double* pval_out = &m_valCrs_Blk[jpsup1*BlkSize];
				for(unsigned int idof=0;idof<BlkSize;idof++){ pval_out[idof] += pval_in[idof]; }
			}
		}
	}
	return true;
}

// bc_flagOłȂRxɌŒ苫EZbg
bool CMatDia_BlkCrs::SetBoundaryCondition(const CBCFlag& bc_flag){

	assert( LenBlkCol() == LenBlkRow() );
	assert( NBlkMatCol() == NBlkMatRow() );
	assert( bc_flag.NBlk() == NBlkMatRow() );
	assert( bc_flag.LenBlk() == LenBlkRow() );

	const unsigned int BlkSize = LenBlkCol()*LenBlkRow();
	const unsigned int BlkLen = LenBlkCol();
	
	// bc_flagOłȂs̍sOɂB
	// AΊp͂PɂĂ
	for(unsigned int iblk=0;iblk<NBlkMatRow();iblk++){
		for(unsigned int idof=0;idof<BlkLen;idof++){
			if( bc_flag.GetBCFlag(iblk,idof) == 0 ) continue;
			{
				for(unsigned int kdof=0;kdof<BlkLen;kdof++){ m_valDia_Blk[iblk*BlkSize+kdof*BlkLen+idof] = 0.0; }
				for(unsigned int jdof=0;jdof<BlkLen;jdof++){ m_valDia_Blk[iblk*BlkSize+idof*BlkLen+jdof] = 0.0; }
				m_valDia_Blk[iblk*BlkSize+idof*BlkLen+idof] = 1.0;
			}
			for(unsigned int ipsup=m_colInd_Blk[iblk];ipsup<m_colInd_Blk[iblk+1];ipsup++){
				for(unsigned int jdof=0;jdof<BlkLen;jdof++){ m_valCrs_Blk[ipsup*BlkSize+idof*BlkLen+jdof] = 0.0; }
			}
		}
	}
	// bc_flagOłȂs̗OɂB
	for(unsigned int ipsup=0;ipsup<this->NCrs();ipsup++){
		const int jblk1 = m_rowPtr_Blk[ipsup];
		for(unsigned int jdof=0;jdof<BlkLen;jdof++){
			if( bc_flag.GetBCFlag(jblk1,jdof) == 0 ) continue;
			for(unsigned int idof=0;idof<BlkLen;idof++){
				m_valCrs_Blk[ipsup*BlkSize+idof*BlkLen+jdof] = 0.0;
			}
		}
	}
	return true;
}

// [p^[
bool CMatDia_BlkCrs::AddPattern(const field::CElemAry& ea){
	if( m_ncrs_Blk == 0 ){
		CMatDia_BlkCrs::DeletePattern();
		ea.MakePointSurPoint(NBlkMatRow(),m_colInd_Blk,m_ncrs_Blk,m_rowPtr_Blk);
	}
	else{
		std::cout << "Error!-->Not Implimentated" << std::endl;
		assert(0);
		abort();
	}
	return true;
}

// [p^[
bool CMatDia_BlkCrs::AddPattern(const CMatDia_BlkCrs& rhs, const bool isnt_trans){
	if(  isnt_trans ){
		assert( NBlkMatCol() == rhs.NBlkMatRow() );
		assert( NBlkMatRow() == rhs.NBlkMatCol() );
		if( m_ncrs_Blk == 0 ){
			DeletePattern();
			CMat_BlkCrs::AddPattern(rhs,isnt_trans);
		}
		else{
			std::cout << "Error!-->Not Implimented" << std::endl;
			assert(0);
			abort();
		}
	}
	else{
		std::cout << "Error!-->Not Implimented" <<  std::endl;
		assert(0);
		abort();
	}
	return true;
}

// [p^[
bool CMatDia_BlkCrs::AddPattern(const CMatDia_BlkCrs& rhs, const COrderTotal_Blk& order){
	assert( NBlkMatCol() == rhs.NBlkMatRow() );
	assert( NBlkMatRow() == rhs.NBlkMatCol() );
	if( m_ncrs_Blk == 0 ){
		DeletePattern();
		CMat_BlkCrs::AddPattern(rhs,order,order);
	}
	else{
		std::cout << "Error!-->Not Implimented" << std::endl;
		assert(0);
		abort();
	}
	return true;
}

bool CMatDia_BlkCrs::AddPattern(const CMat_BlkCrs& m1, const CMatDia_BlkCrs& m2, const CMat_BlkCrs& m3){
	assert( NBlkMatCol()  == m1.NBlkMatCol() );
	assert( m1.NBlkMatRow() == m2.NBlkMatCol() );
	assert( m2.NBlkMatRow() == m3.NBlkMatCol() );
	assert( m3.NBlkMatRow() == NBlkMatRow() );

	if( m_ncrs_Blk == 0 ){

		DeletePattern();

		const unsigned int mat_len_mid = m2.NBlkMatCol();

		m_colInd_Blk[0] = 0;

		std::vector<unsigned int> row_ptr;
		row_ptr.reserve( m2.m_ncrs_Blk*4 );

		////////////////
		unsigned int* is_l_flag = new unsigned int [mat_len_mid];
		for(unsigned int lblk=0;lblk<mat_len_mid;lblk++){ is_l_flag[lblk] = 0; }
		unsigned int l_size = 0;
		unsigned int* l_buffer = new unsigned int [mat_len_mid];
		////////////////
		unsigned int* is_j_flag = new unsigned int [NBlkMatRow()];
		for(unsigned int jblk=0;jblk<NBlkMatRow();jblk++){ is_j_flag[jblk] = 0; }
		unsigned int j_size = 0;
		unsigned int* j_buffer = new unsigned int [NBlkMatRow()];
		////////////////
		for(unsigned int iblk=0;iblk<NBlkMatCol();iblk++){
			{	// Make l_buffer, l_size
				l_size = 0;
				unsigned int npsuk;
				const unsigned int* psuk = m1.GetPtrIndPSuP(iblk,npsuk);
				assert( psuk != 0 );
				for(unsigned int k=0;k<npsuk;k++){
					const unsigned int kblk0 = psuk[k];
					assert( kblk0 < mat_len_mid );
					for(unsigned int klcrs=m2.m_colInd_Blk[kblk0];klcrs<m2.m_colInd_Blk[kblk0+1];klcrs++){
						assert( klcrs < m2.m_ncrs_Blk );
						const unsigned int lblk0 = m2.m_rowPtr_Blk[klcrs];
						assert( lblk0 < mat_len_mid );
						if( is_l_flag[lblk0] == 0 ){
							l_buffer[l_size] = lblk0;
							is_l_flag[lblk0] = l_size+1;
							l_size++;
						}
					}
					// Set Diagonal
					if( is_l_flag[kblk0] == 0 ){
						l_buffer[l_size] = kblk0;
						is_l_flag[kblk0] = l_size+1;
						l_size++;
					}
				}
				for(unsigned int l=0;l<l_size;l++){
					unsigned int lblk0 = l_buffer[l];
					is_l_flag[lblk0] = 0;
				}
			}
			{	// Make j_buffer, j_size
				j_size = 0;
				for(unsigned int l=0;l<l_size;l++){
					unsigned int lblk0 = l_buffer[l];
					unsigned int npsul;
					const unsigned int* psul = m3.GetPtrIndPSuP(lblk0,npsul);
					assert( psul != 0 );
					for(unsigned int j=0;j<npsul;j++){
						const unsigned int jblk0 = psul[j];
						assert( jblk0 < NBlkMatRow() );
						if( jblk0 == iblk ) continue;
						if( is_j_flag[jblk0] == 0 ){
							j_buffer[j_size] = jblk0;
							is_j_flag[jblk0] = j_size+1;
							j_size++;
						}
					}
				}
				for(unsigned int j=0;j<j_size;j++){
					unsigned int jblk0 = j_buffer[j];
					is_j_flag[jblk0] = 0;
				}
			}
			{
				std::sort(&j_buffer[0],&j_buffer[j_size]);
				m_colInd_Blk[iblk+1] = m_colInd_Blk[iblk] + j_size;
				unsigned int ijcrs0 = m_colInd_Blk[iblk];
				row_ptr.resize( row_ptr.size() + j_size );
				for(unsigned int j=0;j<j_size;j++){
					row_ptr[ijcrs0] = j_buffer[j];
					ijcrs0++;
				}
			}
		}
		delete[] l_buffer;	delete[] is_l_flag;	
		delete[] j_buffer;	delete[] is_j_flag; 

		m_ncrs_Blk = row_ptr.size();
		m_rowPtr_Blk = new unsigned int [m_ncrs_Blk];
		for(unsigned int icrs=0;icrs<m_ncrs_Blk;icrs++){
			m_rowPtr_Blk[icrs] = row_ptr[icrs];
		}
	}
	else{
		std::cout << "Error!-->Not Implimented " << std::endl;
		assert(0);
		abort();
	}
 
//	std::cout << "Mat Size : " << m_ncrs_Blk << std::endl;

	return true;
}

bool CMatDia_BlkCrs::SetValue(const CMatDia_BlkCrs& rhs, const COrderTotal_Blk& order)
{
	assert( NBlkMatRow() == rhs.NBlkMatRow() );
	assert( NBlkMatCol() == rhs.NBlkMatCol() );
	assert( NBlkMatCol() == NBlkMatRow() );
	
	assert( LenBlkCol() == rhs.LenBlkCol() );
	assert( LenBlkRow() == rhs.LenBlkRow() );
	assert( LenBlkCol() == LenBlkRow() );

	CMat_BlkCrs::SetValue(rhs,order,order);
	const unsigned int BlkLen = LenBlkCol();
	const unsigned int BlkSize = BlkLen*BlkLen;
	const unsigned int nblk = NBlkMatCol();
	for(unsigned int i0blk=0;i0blk<nblk;i0blk++){
		if( !order.IsInverseMapped(i0blk) ) continue;
		const unsigned int i1blk = (unsigned int)order.InverseMappedBlk(i0blk);
		assert( i1blk < rhs.NBlkMatCol() );
		for(unsigned int idof=0;idof<BlkSize;idof++){
			m_valDia_Blk[i0blk*BlkSize+idof] 
				= rhs.m_valDia_Blk[i1blk*BlkSize+idof];
		}
	}
	return true;
}

bool CMatDia_BlkCrs::SetValue(const CMatDia_BlkCrs& rhs, const bool isnt_trans)
{
	assert( NBlkMatRow() == rhs.NBlkMatRow() );
	assert( NBlkMatCol() == rhs.NBlkMatCol() );
	assert( NBlkMatCol() == NBlkMatRow() );
	
	assert( LenBlkCol() == rhs.LenBlkCol() );
	assert( LenBlkRow() == rhs.LenBlkRow() );
	assert( LenBlkCol() == LenBlkRow() );

	CMat_BlkCrs::SetValue(rhs,isnt_trans);
	if( isnt_trans ){
		const unsigned int ndof=NBlkMatCol()*LenBlkCol()*LenBlkRow();
		for(unsigned int idof=0;idof<ndof;idof++){
			m_valDia_Blk[idof] = rhs.m_valDia_Blk[idof];
		}
	}
	else{
		const unsigned int BlkLen = LenBlkCol();
		const unsigned int BlkSize = BlkLen*BlkLen;
		for(unsigned int iblk=0;iblk<NBlkMatCol();iblk++){
			for(unsigned int idof=0;idof<BlkLen;idof++){
			for(unsigned int jdof=0;jdof<BlkLen;jdof++){
				m_valDia_Blk[iblk*BlkSize+idof*BlkLen+jdof] 
					= rhs.m_valDia_Blk[iblk*BlkSize+jdof*BlkLen*idof];
			}
			}
		}
	}
	return true;
}

bool CMatDia_BlkCrs::SetValue(const CMat_BlkCrs& m1, const CMatDia_BlkCrs& m2, const CMat_BlkCrs& m3){
	assert( NBlkMatRow() == NBlkMatRow() );
	assert( NBlkMatCol() == m1.NBlkMatCol() );
	assert( m1.NBlkMatRow() == m2.NBlkMatCol() );
	assert( m2.NBlkMatRow() == m3.NBlkMatCol() );
	assert( m3.NBlkMatRow() == NBlkMatRow() );

	if( LenBlkCol() != 1 || m2.LenBlkCol() != 1 || LenBlkRow() != 1 ){
		std::cout << "Error!-->Not Implimented " << std::endl;
		assert(0);
		abort();
	}

	const unsigned int mat_len_mid = m2.NBlkMatCol();

	if( m_valCrs_Blk == 0 ){	
		m_valCrs_Blk = new double [m_ncrs_Blk];
	}
	
	////////////////
	unsigned int* is_l_flag = new unsigned int [mat_len_mid];
	for(unsigned int lblk=0;lblk<mat_len_mid;lblk++){ is_l_flag[lblk] = 0; }
	unsigned int l_size = 0;
	unsigned int* l_buffer = new unsigned int [mat_len_mid];
	double* l_val = new double [mat_len_mid];
	////////////////
	unsigned int* is_j_flag = new unsigned int [NBlkMatRow()];
	for(unsigned int jblk=0;jblk<NBlkMatRow();jblk++){ is_j_flag[jblk] = 0; }
	unsigned int j_size = 0;
	unsigned int* j_buffer = new unsigned int [NBlkMatRow()];
	double* j_val = new double [NBlkMatRow()];
	////////////////
	unsigned int* row2crs = new unsigned int [NBlkMatRow()];
	for(unsigned int jblk=0;jblk<NBlkMatRow();jblk++){ 
		row2crs[jblk] = -1; 
	}
	for(unsigned int iblk=0;iblk<NBlkMatCol();iblk++){
		{	// Make l_buffer, l_size
			l_size = 0;
			unsigned int npsuk = 0;
			const unsigned int* ind_psuk = m1.GetPtrIndPSuP(iblk,npsuk);
			assert( ind_psuk != 0 );
			const double* val_psuk = m1.GetPtrValPSuP(iblk,npsuk);
			assert( val_psuk != 0 );
			for(unsigned int k=0;k<npsuk;k++){
				const unsigned int kblk0 = ind_psuk[k];
				assert( kblk0 < mat_len_mid );
				const double& ik_val = val_psuk[k];
				for(unsigned int klcrs=m2.m_colInd_Blk[kblk0];klcrs<m2.m_colInd_Blk[kblk0+1];klcrs++){
					assert( klcrs < m2.m_ncrs_Blk );
					const unsigned int lblk0 = m2.m_rowPtr_Blk[klcrs];
					assert( lblk0 < mat_len_mid );
					if( is_l_flag[lblk0] == 0 ){
						l_buffer[l_size] = lblk0;
						l_val[l_size] = ik_val*m2.m_valCrs_Blk[klcrs];
						is_l_flag[lblk0] = l_size+1;
						l_size++;
					}
					else{
						const unsigned int l = is_l_flag[lblk0]-1;
						assert( l < l_size );
						assert( l_buffer[l] == lblk0 );
						l_val[l] += ik_val*m2.m_valCrs_Blk[klcrs];
					}
				}
				// Set Diagonal
				if( is_l_flag[kblk0] == 0 ){
					l_buffer[l_size] = kblk0;
					l_val[l_size] = ik_val*m2.m_valDia_Blk[kblk0];
					is_l_flag[kblk0] = l_size+1;
					l_size++;
				}
				else{
					const unsigned int l = is_l_flag[kblk0]-1;
					assert(  l_buffer[l] == kblk0 );
					l_val[l] += ik_val*m2.m_valDia_Blk[kblk0];
				}
			}
			for(unsigned int l=0;l<l_size;l++){
				unsigned int lblk0 = l_buffer[l];
				assert( is_l_flag[lblk0] -1 == l );
				is_l_flag[lblk0] = 0;
			}
		}
		{	// Make j_buffer, j_size
			j_size = 0;
			for(unsigned int l=0;l<l_size;l++){
				unsigned int lblk0 = l_buffer[l];
				assert( lblk0  < mat_len_mid );
				const double& lj_val = l_val[l];
				unsigned int npsul = 0;
				const unsigned int* ind_psul = m3.GetPtrIndPSuP(lblk0,npsul);
				assert( ind_psul != 0 );
				const double* val_psul = m3.GetPtrValPSuP(lblk0,npsul);
				assert( val_psul != 0 );
				for(unsigned int l=0;l<npsul;l++){
					const unsigned int jblk0 = ind_psul[l];
					assert( jblk0 < NBlkMatRow() );
					if( is_j_flag[jblk0] == 0 ){
						j_buffer[j_size] = jblk0;
						j_val[j_size] = lj_val*val_psul[l];
						is_j_flag[jblk0] = j_size+1;
						j_size++;
					}
					else{
						const unsigned int j = is_j_flag[jblk0]-1;
						assert( j < j_size );
						assert( j_buffer[j] == jblk0 );
						j_val[j] += lj_val*val_psul[l];
					}
				}
			}
			for(unsigned int j=0;j<j_size;j++){
				unsigned int jblk0 = j_buffer[j];
				assert( is_j_flag[jblk0]-1 == j );
				is_j_flag[jblk0] = 0;
			}
		}
		////////////////
		for(unsigned int ijcrs=m_colInd_Blk[iblk];ijcrs<m_colInd_Blk[iblk+1];ijcrs++){
			assert( ijcrs<m_ncrs_Blk );
			const unsigned int jblk0 = m_rowPtr_Blk[ijcrs];
			assert( jblk0 < NBlkMatRow() );
			row2crs[jblk0] = ijcrs;
		}
		for(unsigned int j=0;j<j_size;j++){
			const unsigned int jblk0 = j_buffer[j];
			assert( jblk0<NBlkMatRow() );
			if( jblk0 != iblk ){
				const int ijcrs0 = row2crs[jblk0];
				if( ijcrs0 == -1 ){	continue; }
				m_valCrs_Blk[ijcrs0] = j_val[j];
			}
			else{
				m_valDia_Blk[iblk] = j_val[j];
			}
		}
		for(unsigned int ijcrs=m_colInd_Blk[iblk];ijcrs<m_colInd_Blk[iblk+1];ijcrs++){
				assert( ijcrs<m_ncrs_Blk );
			const unsigned int jblk0 = m_rowPtr_Blk[ijcrs];
			assert( jblk0 < NBlkMatRow() );
			row2crs[jblk0] = -1;
		}
	}
	delete[] l_buffer;	delete[] is_l_flag;	delete[] l_val;
	delete[] j_buffer;	delete[] is_j_flag; delete[] j_val;
	delete[] row2crs;

	return true;
}

// Calc Matrix Vector Product
// {y} = alpha * [A]{x} + beta * {y}
bool CMatDia_BlkCrs::MatVec(double alpha, const CVector_Blk& x, double beta, CVector_Blk& y) const
{
	assert( NBlkMatCol() == NBlkMatRow() );

	assert( x.BlkVecLen() == NBlkMatRow() );
	assert( x.BlkLen()  == LenBlkRow() );

	assert( y.BlkVecLen() == NBlkMatCol() );
	assert( y.BlkLen()  == LenBlkCol() );

	const unsigned int BlkLen = LenBlkCol();
	const unsigned int BlkSize = BlkLen*BlkLen;

	const double* xval = x.m_Value;
	double* yval = y.m_Value;

	if( BlkLen == 1 ){
		for(unsigned int iblk=0;iblk<NBlkMatCol();iblk++){
			double* iyval = yval+iblk;
			(*iyval) *= beta;
			for(unsigned int icrs=m_colInd_Blk[iblk];icrs<m_colInd_Blk[iblk+1];icrs++){
				assert( icrs < m_ncrs_Blk );
				const unsigned int jblk0 = m_rowPtr_Blk[icrs];
				assert( jblk0 < NBlkMatRow() );
				const double* jxval = xval+jblk0;
				(*iyval) += alpha * m_valCrs_Blk[icrs] * (*jxval);
			}
			const double* ixval = xval + iblk;
			(*iyval) += alpha * m_valDia_Blk[iblk] * (*ixval);
		}
	}
	else if( BlkLen > 1 ){
		for(unsigned int iblk=0;iblk<NBlkMatCol();iblk++){
			double* iyval = yval+iblk*BlkLen;
			{
				for(unsigned int idof=0;idof<BlkLen;idof++){
					iyval[idof] *= beta;
				}
			}
			for(unsigned int icrs=m_colInd_Blk[iblk];icrs<m_colInd_Blk[iblk+1];icrs++){
				assert( icrs < m_ncrs_Blk );
				const unsigned int jblk0 = m_rowPtr_Blk[icrs];
				assert( jblk0 < NBlkMatRow() );
				{
					const double* jxval = xval+jblk0*BlkLen;
					for(unsigned int idof=0;idof<BlkLen;idof++){
					for(unsigned int jdof=0;jdof<BlkLen;jdof++){
						iyval[idof] += alpha * m_valCrs_Blk[icrs*BlkSize+idof*BlkLen+jdof] * jxval[jdof];
					}
					}
				}
			}
			{
				const double* ixval = xval + iblk*BlkLen;
				for(unsigned int idof=0;idof<BlkLen;idof++){
				for(unsigned int jdof=0;jdof<BlkLen;jdof++){
					iyval[idof] += alpha * m_valDia_Blk[iblk*BlkSize+idof*BlkLen+jdof] * ixval[jdof];
				}
				}
			}
		}
	}
	return true;
}
