//
// Mat_Crs.cpp: CMat_BlkCrs NX̃Cve[V
//
//			Copy Rights (c) Nobuyuki Umetani 2007
//          e-mail : umetani@sml.k.u-tokyo.ac.jp 
//			Last Updated 02/21/2007				
//////////////////////////////////////////////////////////////////////

#pragma warning( disable : 4786 )
#define for if(0); else for

#include <cassert>
#include <iostream>
#include <algorithm>

#include "Mat_BlkCrs.h"
#include "Vector_Blk.h"
#include "Order_Blk.h"

using namespace ls;

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

CMat_BlkCrs::CMat_BlkCrs()
: m_nblk_MatCol(0), m_nblk_MatRow(0), m_len_BlkCol(0), m_len_BlkRow(0)
{
//	std::cout << "Construct : CMat_BlkCrs (Default Constructor) " << std::endl;
	m_colInd_Blk = 0;
	m_ncrs_Blk = 0;
	m_rowPtr_Blk = 0;
	m_valCrs_Blk = 0;
}

CMat_BlkCrs::CMat_BlkCrs(const unsigned int nblk_col, const unsigned int len_col, 
				   const unsigned int nblk_row, const unsigned int len_row )
: m_nblk_MatCol(nblk_col), m_nblk_MatRow(nblk_row), m_len_BlkCol(len_col), m_len_BlkRow(len_row)
{
//	std::cout << "Construct : CMat_BlkCrs " << std::endl;
	m_colInd_Blk = new unsigned int [m_nblk_MatCol+1];
	for(unsigned int iblk=0;iblk<m_nblk_MatCol+1;iblk++){ m_colInd_Blk[iblk] = 0; }
	m_ncrs_Blk = 0;
	m_rowPtr_Blk = 0;
	m_valCrs_Blk = 0; 
}

CMat_BlkCrs::CMat_BlkCrs(const CMat_BlkCrs& rhs, const bool is_value, const bool isnt_trans)
: m_nblk_MatCol( isnt_trans ? rhs.m_nblk_MatCol : rhs.m_nblk_MatRow ),
m_nblk_MatRow( isnt_trans ? rhs.m_nblk_MatRow : rhs.m_nblk_MatCol ),
m_len_BlkCol( isnt_trans ? rhs.m_len_BlkCol : rhs.m_len_BlkRow ),
m_len_BlkRow( isnt_trans ? rhs.m_len_BlkRow : rhs.m_len_BlkCol )
{
//	std::cout << "Construct : CMat_BlkCrs" << std::endl;
	m_colInd_Blk = new unsigned int [m_nblk_MatCol+1];
	for(unsigned int iblk=0;iblk<m_nblk_MatCol+1;iblk++){ m_colInd_Blk[iblk] = 0; }
	m_ncrs_Blk = 0;
	m_rowPtr_Blk = 0;
	m_valCrs_Blk = 0; 
	AddPattern(rhs,isnt_trans);
	if( is_value ){ this->SetValue(rhs,isnt_trans); }
}

void CMat_BlkCrs::Initialize(const unsigned int nblk_col, const unsigned int len_col, 
						 const unsigned int nblk_row, const unsigned int len_row ){

	if( m_colInd_Blk != 0 ){ delete[] m_colInd_Blk; }
	if( m_rowPtr_Blk != 0 ){ delete[] m_rowPtr_Blk; }
	if( m_valCrs_Blk != 0 ){ delete[] m_valCrs_Blk; }

	m_nblk_MatCol = nblk_col;
	m_nblk_MatRow = nblk_row;
	m_len_BlkCol = len_col;
	m_len_BlkRow = len_row;

	m_colInd_Blk = new unsigned int [m_nblk_MatCol+1];
	for(unsigned int iblk=0;iblk<m_nblk_MatCol+1;iblk++){ m_colInd_Blk[iblk] = 0; }
	m_ncrs_Blk = 0;
	m_rowPtr_Blk = 0;
	m_valCrs_Blk = 0; 
}


CMat_BlkCrs::~CMat_BlkCrs()
{
	if( m_colInd_Blk != 0 ){ delete[] m_colInd_Blk;	m_colInd_Blk = 0; }
	if( m_rowPtr_Blk != 0 ){ delete[] m_rowPtr_Blk;	m_rowPtr_Blk = 0; }
	if( m_valCrs_Blk != 0 ){ delete[] m_valCrs_Blk;	m_valCrs_Blk = 0; }
}

bool CMat_BlkCrs::DeletePattern(){
	m_ncrs_Blk = 0;
	for(unsigned int iblk=0;iblk<m_nblk_MatCol+1;iblk++){ m_colInd_Blk[iblk] = 0; }
	if( m_rowPtr_Blk != 0 ){ delete[] m_rowPtr_Blk; m_rowPtr_Blk = 0; }
	if( m_valCrs_Blk != 0 ){ delete[] m_valCrs_Blk; m_valCrs_Blk = 0; }
	return true;
}

bool CMat_BlkCrs::SetZero(){
	const unsigned int ntotdof = m_ncrs_Blk*m_len_BlkCol*m_len_BlkRow;
	if( m_valCrs_Blk == 0 ){	// ܂l̗̈mۂĂȂꍇ
		m_valCrs_Blk = new double [ntotdof];
	}
	for(unsigned int idof=0;idof<ntotdof;idof++){ m_valCrs_Blk[idof] = 0.0;	}
	return true;
}

bool CMat_BlkCrs::AddPattern(const CMat_BlkCrs& rhs, const bool isnt_trans)
{
	if( rhs.m_ncrs_Blk == 0 ) return true;
	if( isnt_trans ){	// Add Not Transpose Pattern of rhs
		if( this->m_ncrs_Blk == 0 ){
			if( this->m_rowPtr_Blk != 0 ){ delete[] this->m_rowPtr_Blk; }
			assert( m_nblk_MatCol == rhs.m_nblk_MatCol );
			assert( m_nblk_MatRow == rhs.m_nblk_MatRow );
			for(unsigned int iblk=0;iblk<m_nblk_MatCol+1;iblk++){
				this->m_colInd_Blk[iblk] = rhs.m_colInd_Blk[iblk];
			}
			m_ncrs_Blk = m_colInd_Blk[m_nblk_MatCol];
			assert( m_ncrs_Blk == rhs.m_ncrs_Blk );
			
			if( m_rowPtr_Blk != 0 ){ delete[] m_rowPtr_Blk; m_rowPtr_Blk = 0; }
			this->m_ncrs_Blk = rhs.m_ncrs_Blk;
			this->m_rowPtr_Blk = new unsigned int [m_ncrs_Blk];
			for(unsigned int icrs=0;icrs<m_ncrs_Blk;icrs++){
				this->m_rowPtr_Blk[icrs] = rhs.m_rowPtr_Blk[icrs];
			}
		}
		else{
			std::cout << "Error!-->Not Implimented" << std::endl;
			assert(0);
			abort();
			/* include 邩ׂĂAincludeȂȂp^[t */
		}
	}
	else{	// Add Transpose Pattern of rhs
		if( this->m_ncrs_Blk == 0 ){
			if( this->m_rowPtr_Blk != 0 ){ delete[] this->m_rowPtr_Blk; }
			assert( m_nblk_MatCol == rhs.m_nblk_MatRow );
			assert( m_nblk_MatRow == rhs.m_nblk_MatCol );
			for(unsigned int iblk=0;iblk<m_nblk_MatCol+1;iblk++){
				m_colInd_Blk[iblk] = 0;
			}
			for(unsigned int jblk=0;jblk<m_nblk_MatRow;jblk++){
				for(unsigned int jicrs=rhs.m_colInd_Blk[jblk];jicrs<rhs.m_colInd_Blk[jblk+1];jicrs++){
					assert( jicrs < rhs.m_ncrs_Blk );
					const unsigned int iblk0 = rhs.m_rowPtr_Blk[jicrs];
					assert( iblk0 < m_nblk_MatCol);
					m_colInd_Blk[iblk0+1]++;
				}
			}
			for(unsigned int iblk=1;iblk<m_nblk_MatCol+1;iblk++){
				m_colInd_Blk[iblk] += m_colInd_Blk[iblk-1];
			}
			m_ncrs_Blk = m_colInd_Blk[m_nblk_MatCol];
			assert( m_ncrs_Blk == rhs.m_ncrs_Blk );
			if( m_rowPtr_Blk != 0 ){ delete[] m_rowPtr_Blk; m_rowPtr_Blk = 0; }
			m_rowPtr_Blk = new unsigned int [m_ncrs_Blk];
			for(unsigned int jblk=0;jblk<m_nblk_MatRow;jblk++){
				for(unsigned int jicrs=rhs.m_colInd_Blk[jblk];jicrs<rhs.m_colInd_Blk[jblk+1];jicrs++){
					assert( jicrs < rhs.m_ncrs_Blk );
					const unsigned int iblk0 = rhs.m_rowPtr_Blk[jicrs];
					assert( iblk0 < m_nblk_MatCol );
					const unsigned int ijcrs = m_colInd_Blk[iblk0];
					assert( ijcrs < m_ncrs_Blk );
					m_rowPtr_Blk[ijcrs] = jblk;
					m_colInd_Blk[iblk0]++;
				}
			}
			for(unsigned int iblk=m_nblk_MatCol;iblk>=1;iblk--){
				m_colInd_Blk[iblk] = m_colInd_Blk[iblk-1];
			}
			m_colInd_Blk[0] = 0;
		}
		else{
			std::cout << "Error!-->Not Implimented" << std::endl;
			assert(0);
			abort();
			/* include 邩ׂĂAincludeȂȂp^[t */
		}
	}
	return true;
}

bool CMat_BlkCrs::AddPattern(const CMat_BlkCrs& rhs, 
		const COrderTotal_Blk& order_col, const COrderTotal_Blk& order_row)
{
	assert( rhs.NBlkMatCol() == order_col.NBlkFrom() );
	assert( rhs.NBlkMatRow() == order_row.NBlkFrom() );
	assert( this->NBlkMatCol() == order_col.NBlkTo() );
	assert( this->NBlkMatRow() == order_row.NBlkTo() );
	assert( this->LenBlkCol() == rhs.LenBlkCol() );
	assert( this->LenBlkRow() == rhs.LenBlkRow() );

	if( rhs.m_ncrs_Blk == 0 ) return true;
	if( this->m_ncrs_Blk == 0 ){
		if( m_rowPtr_Blk != 0 ){ delete[] m_rowPtr_Blk; m_rowPtr_Blk = 0; }
		assert( m_nblk_MatCol == rhs.m_nblk_MatRow );
		assert( m_nblk_MatRow == rhs.m_nblk_MatCol );
		for(unsigned int i0blk=0;i0blk<m_nblk_MatCol+1;i0blk++){
			m_colInd_Blk[i0blk] = 0;
		}
		for(unsigned int i0blk=0;i0blk<m_nblk_MatCol;i0blk++){
			if( !order_col.IsInverseMapped(i0blk) ) continue;
			const unsigned int i1blk0 = order_col.InverseMappedBlk(i0blk);
			for(unsigned int ij1crs=rhs.m_colInd_Blk[i1blk0];ij1crs<rhs.m_colInd_Blk[i1blk0+1];ij1crs++){
				assert( ij1crs < rhs.m_ncrs_Blk );
				const unsigned int j1blk0 = rhs.m_rowPtr_Blk[ij1crs];
				assert( j1blk0 < rhs.NBlkMatRow() );
				if( !order_row.IsMapped(j1blk0) ) continue;
				m_colInd_Blk[i0blk+1]++;
			}
		}
		for(unsigned int i0blk=1;i0blk<m_nblk_MatCol+1;i0blk++){
			m_colInd_Blk[i0blk] += m_colInd_Blk[i0blk-1];
		}
		m_ncrs_Blk = m_colInd_Blk[m_nblk_MatCol];
		assert( m_ncrs_Blk == rhs.m_ncrs_Blk );
		m_rowPtr_Blk = new unsigned int [m_ncrs_Blk];
		for(unsigned int i0blk=0;i0blk<m_nblk_MatCol;i0blk++){
			if( !order_col.IsInverseMapped(i0blk) ) continue;
			const unsigned int i1blk0 = order_col.InverseMappedBlk(i0blk);
			for(unsigned int ij1crs=rhs.m_colInd_Blk[i1blk0];ij1crs<rhs.m_colInd_Blk[i1blk0+1];ij1crs++){
				assert( ij1crs < rhs.m_ncrs_Blk );
				const unsigned int j1blk0 = rhs.m_rowPtr_Blk[ij1crs];
				assert( j1blk0 < rhs.NBlkMatRow() );
				if( !order_row.IsMapped(j1blk0) ) continue;
				const unsigned int j0blk0 = (unsigned int)order_row.MappedBlk(j1blk0);
				assert( j0blk0 < m_nblk_MatRow );
				const unsigned int ij0crs0 = m_colInd_Blk[i0blk];
				assert( ij0crs0 < m_ncrs_Blk );
				m_rowPtr_Blk[ij0crs0] = j0blk0;
				m_colInd_Blk[i0blk]++;
			}
		}
		for(unsigned int iblk=m_nblk_MatCol;iblk>=1;iblk--){
			m_colInd_Blk[iblk] = m_colInd_Blk[iblk-1];
		}
		m_colInd_Blk[0] = 0;
		for(unsigned int iblk=0;iblk<m_nblk_MatCol;iblk++){
			const unsigned int ijcrs0 = m_colInd_Blk[iblk  ];
			const unsigned int ijcrs1 = m_colInd_Blk[iblk+1];
			std::sort(&m_rowPtr_Blk[ijcrs0],&m_rowPtr_Blk[ijcrs1]);
		}
	}
	else{
		std::cout << "Error!-->Not Implimented" << std::endl;
		assert(0);
		abort();
		/* include 邩ׂĂAincludeȂȂp^[t */
	}
	return true;
}

bool CMat_BlkCrs::SetValue(const CMat_BlkCrs& rhs, const bool isnt_trans){
	const unsigned int BlkSize = m_len_BlkCol*m_len_BlkRow;
	this->SetZero();
	if( isnt_trans ){	// Set Value -- Not Transpose Pattern
		assert( m_nblk_MatCol == rhs.m_nblk_MatCol );
		assert( m_nblk_MatRow == rhs.m_nblk_MatRow );
		assert( m_len_BlkCol == rhs.m_len_BlkCol );
		assert( m_len_BlkRow == rhs.m_len_BlkRow );
		unsigned int* row2crs = new unsigned int [m_nblk_MatRow];
		for(unsigned int jblk=0;jblk<m_nblk_MatRow;jblk++){ row2crs[jblk] = -1; }	
		for(unsigned int iblk=0;iblk<m_nblk_MatCol;iblk++){
			// iblks̃p^[񂩂crsԍւ̃nbV
			for(unsigned int ijcrs=m_colInd_Blk[iblk];ijcrs<m_colInd_Blk[iblk+1];ijcrs++){
				assert( ijcrs<m_ncrs_Blk );
				const unsigned int jblk0 = m_rowPtr_Blk[ijcrs];
				assert( jblk0 < m_nblk_MatRow );
				row2crs[jblk0] = ijcrs;
			}
			for(unsigned int ijcrs=rhs.m_colInd_Blk[iblk];ijcrs<rhs.m_colInd_Blk[iblk+1];ijcrs++){
				assert( ijcrs<rhs.m_ncrs_Blk );
				const unsigned int jblk0 = rhs.m_rowPtr_Blk[ijcrs];
				assert( jblk0<m_nblk_MatRow );
				const int ijcrs0 = row2crs[jblk0];
				if( ijcrs0 == -1 ) continue;
				////////////////
				const double* pval_in = &rhs.m_valCrs_Blk[ijcrs*BlkSize];
				double* pval_out = &m_valCrs_Blk[ijcrs0*BlkSize];
				for(unsigned int idof=0;idof<BlkSize;idof++){ *(pval_out+idof) = *(pval_in+idof); }
			}
			// nbV̏
			for(unsigned int ijcrs=m_colInd_Blk[iblk];ijcrs<m_colInd_Blk[iblk+1];ijcrs++){
				assert( ijcrs<m_ncrs_Blk );
				const unsigned int jblk0 = m_rowPtr_Blk[ijcrs];
				assert( jblk0 < m_nblk_MatRow );
				row2crs[jblk0] = -1;
			}
		}
		delete[] row2crs;
	}
	else{	// Set Value -- Transpose Pattern
		assert( m_nblk_MatCol == rhs.m_nblk_MatRow );
		assert( m_nblk_MatRow == rhs.m_nblk_MatCol );
		assert( m_len_BlkCol == rhs.m_len_BlkRow );
		assert( m_len_BlkRow == rhs.m_len_BlkCol );
		for(unsigned int iblk=0;iblk<m_nblk_MatCol;iblk++){
			for(unsigned int ijcrs=m_colInd_Blk[iblk];ijcrs<m_colInd_Blk[iblk+1];ijcrs++){
				assert( ijcrs<m_ncrs_Blk );
				const unsigned int jblk0 = m_rowPtr_Blk[ijcrs];
				assert( jblk0 < m_nblk_MatRow );
				const unsigned int jicrs_begin = rhs.m_colInd_Blk[jblk0];
				const unsigned int jicrs_end = rhs.m_colInd_Blk[jblk0+1];
				if( jicrs_begin == jicrs_end ) continue;
				if( iblk<rhs.m_rowPtr_Blk[jicrs_begin] ) continue;
				if( rhs.m_rowPtr_Blk[ jicrs_end-1 ] < iblk  ) continue;
				for(unsigned int jicrs=rhs.m_colInd_Blk[jblk0];jicrs<rhs.m_colInd_Blk[jblk0+1];jicrs++){
					assert( jicrs<rhs.m_ncrs_Blk );
					const unsigned int iblk0 = rhs.m_rowPtr_Blk[jicrs];
					if( iblk != iblk0 ) continue;
					////////////////
					const double* pval_in = &rhs.m_valCrs_Blk[jicrs*BlkSize];
					double* pval_out = &m_valCrs_Blk[ijcrs*BlkSize];
					for(unsigned int idof=0;idof<m_len_BlkCol;idof++){
						for(unsigned int jdof=0;jdof<m_len_BlkRow;jdof++){
							*(pval_out+idof*m_len_BlkRow+jdof) = *(pval_in+jdof*m_len_BlkCol+idof);
						}
					}
					////////////////
				}
			}
		}
	}
	return true;
}

bool CMat_BlkCrs::SetValue(const CMat_BlkCrs& rhs, 
		const COrderTotal_Blk& order_col, const COrderTotal_Blk& order_row){

	assert( rhs.NBlkMatCol() == order_col.NBlkFrom() );
	assert( rhs.NBlkMatRow() == order_row.NBlkFrom() );
	assert( this->NBlkMatCol() == order_col.NBlkTo() );
	assert( this->NBlkMatRow() == order_row.NBlkTo() );
	assert( this->LenBlkCol() == rhs.LenBlkCol() );
	assert( this->LenBlkRow() == rhs.LenBlkRow() );

	const unsigned int BlkSize = m_len_BlkCol*m_len_BlkRow;
	this->SetZero();

	unsigned int* row2crs = new unsigned int [m_nblk_MatRow];
	for(unsigned int jblk=0;jblk<m_nblk_MatRow;jblk++){ 
		row2crs[jblk] = -1; 
	}
	for(unsigned int i0blk=0;i0blk<m_nblk_MatCol;i0blk++){
		if( !order_col.IsInverseMapped(i0blk) ) continue;
		for(unsigned int ij0crs=m_colInd_Blk[i0blk];ij0crs<m_colInd_Blk[i0blk+1];ij0crs++){
			assert( ij0crs<m_ncrs_Blk );
			const unsigned int j0blk0 = m_rowPtr_Blk[ij0crs];
			assert( j0blk0 < m_nblk_MatRow );
			row2crs[j0blk0] = ij0crs;
		}
		const unsigned int i1blk0 = (unsigned int)order_col.InverseMappedBlk(i0blk);
		for(unsigned int ij1crs=rhs.m_colInd_Blk[i1blk0];ij1crs<rhs.m_colInd_Blk[i1blk0+1];ij1crs++){
			assert( ij1crs<rhs.m_ncrs_Blk );
			const unsigned int j1blk0 = rhs.m_rowPtr_Blk[ij1crs];
			assert( j1blk0<rhs.NBlkMatRow() );
			if( !order_row.IsMapped(j1blk0) ) continue;
			////////////////
			const unsigned int j0blk0 = order_row.MappedBlk(j1blk0);
			assert( j0blk0<m_nblk_MatRow );
			const int ij0crs0 = row2crs[j0blk0];
			if( ij0crs0 == -1 ) continue;
			////////////////
			const double* pval_in = &rhs.m_valCrs_Blk[ij1crs*BlkSize];
			double* pval_out = &m_valCrs_Blk[ij0crs0*BlkSize];
			for(unsigned int idof=0;idof<BlkSize;idof++){
				*(pval_out+idof) = *(pval_in+idof);
			}
		}
		for(unsigned int ij0crs=m_colInd_Blk[i0blk];ij0crs<m_colInd_Blk[i0blk+1];ij0crs++){
			assert( ij0crs<m_ncrs_Blk );
			const unsigned int j0blk0 = m_rowPtr_Blk[ij0crs];
			assert( j0blk0 < m_nblk_MatRow );
			row2crs[j0blk0] = -1;
		}
	}
	delete[] row2crs;
	return true;
}

bool CMat_BlkCrs::MatVec(double alpha, const CVector_Blk& x, double beta, CVector_Blk& b, const bool isnt_trans) const
{
	////////////////
	if( this->m_len_BlkCol != 1 ){
		std::cout << "Error!-->" << std::endl;
		assert(0);
	}
	////////////////

	assert( x.BlkVecLen() == m_nblk_MatRow );
	assert( x.BlkLen() == m_len_BlkRow );
	assert( b.BlkVecLen() == m_nblk_MatCol );
	assert( b.BlkLen() == m_len_BlkCol );

	if( isnt_trans ){
		if( ((beta>=0.0)?beta:-beta) > 1.0e-30 ){
			for(unsigned int iblk=0;iblk<m_nblk_MatCol;iblk++){
				double* lval = &(b.m_Value[iblk]);
				*lval *= beta;
				for(unsigned int icrs=m_colInd_Blk[iblk];icrs<m_colInd_Blk[iblk+1];icrs++){
					assert( icrs < m_ncrs_Blk );
					const unsigned int jblk0 = m_rowPtr_Blk[icrs];
					assert( jblk0 < m_nblk_MatRow );
					*lval += alpha * m_valCrs_Blk[icrs]*x.m_Value[jblk0];
				}
			}
		}
		else{
			for(unsigned int iblk=0;iblk<m_nblk_MatCol;iblk++){
				double* lval = &(b.m_Value[iblk]);
				*lval = 0.0;
				for(unsigned int icrs=m_colInd_Blk[iblk];icrs<m_colInd_Blk[iblk+1];icrs++){
					assert( icrs < m_ncrs_Blk );
					const unsigned int jblk0 = m_rowPtr_Blk[icrs];
					assert( jblk0 < m_nblk_MatRow );
					*lval += alpha * m_valCrs_Blk[icrs]*x.m_Value[jblk0];
				}
			}
		}
	}
	else{
		for(unsigned int jblk=0;jblk<m_nblk_MatRow;jblk++){
			b.m_Value[jblk] *= beta;
		}
		for(unsigned int iblk=0;iblk<m_nblk_MatCol;iblk++){
			for(unsigned int icrs=m_colInd_Blk[iblk];icrs<m_colInd_Blk[iblk+1];icrs++){
				assert( icrs < m_ncrs_Blk );
				const unsigned int jblk0 = m_rowPtr_Blk[icrs];
				assert( jblk0 < m_nblk_MatRow );
				b.m_Value[jblk0] += alpha * m_valCrs_Blk[icrs]*x.m_Value[iblk];
			}
		}
	}
	return true;
}

