// Mat_Crs.h: CMat_BlkCrs NX̃C^[tFCX
//
//			Copy Rights (c) Nobuyuki Umetani 2007
//          e-mail : umetani@sml.k.u-tokyo.ac.jp 
//			Last Updated 15/04/2007		
//////////////////////////////////////////////////////////////////////

#if !defined(MAT_CRS_H)
#define MAT_CRS_H

#include <assert.h>

namespace ls{

class CVector_Blk;
class CMatDia_Crs;
class COrderTotal_Blk;

class CMat_BlkCrs  
{
public:
	CMat_BlkCrs(const unsigned int nblk_col, const unsigned int len_col, 
		const unsigned int nblk_row, const unsigned int len_row );
	CMat_BlkCrs(const CMat_BlkCrs& rhs, const bool is_value, const bool is_trans);
	CMat_BlkCrs(); // Default Constructor
	virtual ~CMat_BlkCrs();

	void Initialize(const unsigned int nblk_col, const unsigned int len_col, 
		const unsigned int nblk_row, const unsigned int len_row );

	inline unsigned int NBlkMatCol() const { return m_nblk_MatCol; }
	inline unsigned int NBlkMatRow() const { return m_nblk_MatRow; }
	
	inline unsigned int LenBlkCol() const { return m_len_BlkCol; }
	inline unsigned int LenBlkRow() const { return m_len_BlkRow; }

	virtual bool SetZero();
	virtual bool DeletePattern();
	
	virtual bool AddPattern(const CMat_BlkCrs& rhs, const bool isnt_trans);
	virtual bool AddPattern(const CMat_BlkCrs& rhs, const COrderTotal_Blk& order_col_rhs2this, const COrderTotal_Blk& order_row_rhs2this);

	virtual bool SetValue(  const CMat_BlkCrs& rhs, const bool isnt_trans);
	virtual bool SetValue(  const CMat_BlkCrs& rhs, const COrderTotal_Blk& order_col_rhs2this, const COrderTotal_Blk& order_row_rhs2this);
	
	virtual bool MatVec(double alpha, const CVector_Blk& x, double beta, CVector_Blk& b, const bool isnt_trans) const;

	////////////////////////////////////////////////
	// Crs Original Function ( Non-virtual )

	const unsigned int NCrs() const { return m_ncrs_Blk; }
	const unsigned int* GetPtrIndPSuP(const unsigned int ipoin, unsigned int& npsup) const
	{
		if( m_rowPtr_Blk == 0 ){ npsup = 0; return 0; }
		assert( ipoin < m_nblk_MatCol );
		npsup = m_colInd_Blk[ipoin+1]-m_colInd_Blk[ipoin];
		const unsigned int icrs = m_colInd_Blk[ipoin];
		return &m_rowPtr_Blk[icrs];
	}
	const double* GetPtrValPSuP(const unsigned int ipoin, unsigned int& npsup) const
	{
		if( m_rowPtr_Blk == 0 ){ npsup = 0; return 0; }
		assert( ipoin < m_nblk_MatCol );
		npsup = m_colInd_Blk[ipoin+1]-m_colInd_Blk[ipoin];
		const unsigned int icrs = m_colInd_Blk[ipoin];
		return &m_valCrs_Blk[icrs];
	}

private:
	unsigned int m_nblk_MatCol;
	unsigned int m_nblk_MatRow;

	unsigned int m_len_BlkCol;
	unsigned int m_len_BlkRow;

protected:
	unsigned int  m_ncrs_Blk;
	unsigned int* m_colInd_Blk;
	unsigned int* m_rowPtr_Blk;

	double* m_valCrs_Blk;
};

}	// end name space 'ls'

#endif 
