// SubVec.cpp: CSubVec NX̃Cve[V
//
//			Copy Rights (c) Nobuyuki Umetani 2007
//          e-mail : umetani@sml.k.u-tokyo.ac.jp 
//			Last Updated 21/Feb/2007		
//////////////////////////////////////////////////////////////////////

#include <cassert>
#include <iostream>

#include "Vector_Blk.h"

using namespace ls;

//////////////////////////////////////////////////////////////////////
// 񃁃õth̃Iy[^
//////////////////////////////////////////////////////////////////////

namespace ls{

double operator*(const CVector_Blk& lhs, const CVector_Blk& rhs){	
	assert( lhs.m_BlkLen == rhs.m_BlkLen );
	assert( lhs.m_BlkVecLen == rhs.m_BlkVecLen );
	const int ndof = lhs.m_BlkLen * lhs.m_BlkVecLen;
	double d = 0.0;
	for(int idof=0;idof<ndof;idof++){ d+=lhs.m_Value[idof]*rhs.m_Value[idof]; }
	return d;
}

}

////////////////////////////////////////////////
// o[Iy[^
////////////////////////////////////////////////

CVector_Blk& CVector_Blk::operator*=(double d0){	// Scaler Product
	const int ndof = m_BlkLen * m_BlkVecLen;
	for(int idof=0;idof<ndof;idof++){ m_Value[idof] *= d0; }
	return *this; 
}

CVector_Blk& CVector_Blk::operator=(const CVector_Blk& rhs){ // Substitue Vector
	assert( m_BlkLen == rhs.m_BlkLen ); 
	assert( m_BlkVecLen == rhs.m_BlkVecLen );
	const int ndof = m_BlkLen*m_BlkVecLen;
	for(int idof=0;idof<ndof;idof++){ m_Value[idof] = rhs.m_Value[idof]; }
	return *this; 
}

CVector_Blk& CVector_Blk::operator+=(const CVector_Blk& rhs){ // Add 
	assert( m_BlkLen == rhs.m_BlkLen ); 
	assert( m_BlkVecLen == rhs.m_BlkVecLen );
	const int ndof = m_BlkLen*m_BlkVecLen;
	for(int idof=0;idof<ndof;idof++){ m_Value[idof] += rhs.m_Value[idof]; }
	return *this; 
}

////////////////////////////////////////////////
// o֐
////////////////////////////////////////////////

CVector_Blk& CVector_Blk::AXPY(const double& alpha, const CVector_Blk& rhs){ // Add scaler scaled Vector
	assert( m_BlkLen == rhs.m_BlkLen ); 
	assert( m_BlkVecLen == rhs.m_BlkVecLen );
	const int ndof = m_BlkLen * m_BlkVecLen;
	for(int idof=0;idof<ndof;idof++){ m_Value[idof] += alpha * rhs.m_Value[idof]; }
	return *this;
}


void CVector_Blk::SetVectorZero(){	// Set 0 to Value
	const unsigned int ndof = m_BlkLen * m_BlkVecLen;
	for(unsigned int idof=0;idof<ndof;idof++){ m_Value[idof]=0.0; } 
}

double CVector_Blk::GetSquaredVectorNorm() const{
	const unsigned int ndof = m_BlkLen * m_BlkVecLen;
	double dtmp1 = 0.0;
	for(unsigned int idof=0;idof<ndof;idof++){ 
		dtmp1 += m_Value[idof]*m_Value[idof]; 
	}
	return dtmp1;
}
