// Vec.h: CVector_Blk NX̃C^[tFCX
//
//			Copy Rights (c) Nobuyuki Umetani 2007
//          e-mail : umetani@sml.k.u-tokyo.ac.jp 
//			Last Updated 07.May.2007		
//////////////////////////////////////////////////////////////////////

#if !defined(VEC_H)
#define VEC_H

#include <assert.h>

namespace ls {

class CMat_BlkCrs;
class CMatDia_BlkCrs;
class CVector_Blk  
{
	friend class CMat_BlkCrs;
	friend class CMatDia_BlkCrs;
	friend class CMatDiaFrac_BlkCrs;
	friend class CMatDiaInv_BlkDia;
    friend double operator*(const CVector_Blk& lhs, const CVector_Blk& rhs);	// Dot Product
public:
	CVector_Blk(const unsigned int& iblkveclen, const unsigned int& iblklen) 
		: m_BlkVecLen(iblkveclen), m_BlkLen(iblklen){
		m_Value = new double [m_BlkLen*m_BlkVecLen];
	}
	virtual ~CVector_Blk(){ if( m_Value!=0) delete[] m_Value; }

	CVector_Blk& operator*=(double d0);	// Scaler Product
	CVector_Blk& operator=(const CVector_Blk& rhs); // Substitue Vector
	CVector_Blk& operator+=(const CVector_Blk& rhs); // Add 


	////////////////////////////////
	// Menber function

	CVector_Blk& AXPY(const double& alpha, const CVector_Blk& rhs); // Add scaler scaled Vector

	void SetVectorZero();	// Set 0 to Value
	double GetSquaredVectorNorm() const;

	unsigned int BlkLen() const { return m_BlkLen; }	// Size of One Block
	unsigned int BlkVecLen() const { return m_BlkVecLen; }	// Size of Bocks Vector have

	const double& GetValue(const unsigned int& iblk, const unsigned int& idofblk) const{
		assert( iblk<m_BlkVecLen ); 
		assert( idofblk<m_BlkLen );
		return m_Value[iblk*m_BlkLen+idofblk]; 
	}
	void SetValue(const unsigned int& iblk, const unsigned int& idofblk, const double& val){
		assert( iblk<m_BlkVecLen ); assert( idofblk<m_BlkLen );
		m_Value[iblk*m_BlkLen+idofblk] = val; 	
	}
	void AddValue(const unsigned int& iblk, const unsigned int& idofblk, const double& val){
		assert( iblk<m_BlkVecLen ); assert( idofblk<m_BlkLen );
		m_Value[iblk*m_BlkLen+idofblk] += val; 	
	}
private:
	const unsigned int m_BlkLen;
	const unsigned int m_BlkVecLen;
	double* m_Value;
};

}	// end namespace 'ls'

#endif // VEC_H
