////////////////////////////////////////////////////////////////
//															  //
//	  Linear Elastic FEM Code ( Unstructed Grid ) ..ver 0.0.0 //
//															  //
//			Copy Rights (c) Nobuyuki Umetani 2007			  //
//          e-mail : n.umetani@gmai.com                       //
//			Last Updated 20/July/2007						  //
////////////////////////////////////////////////////////////////

#pragma warning( disable : 4786 )
#define for if(0); else for

#include <cassert>
#include <iostream>
#include <string>
#include <vector>
#include <set>
#include <math.h>

#include "Field/File.h"
#include "Field/ElemAry_UNS.h"
#include "Field/NodeAry.h"
#include "Field/ElemArySet.h"
#include "Field/Field.h"

#include "Mat/MatDia_BlkCrs.h"
#include "Mat/MatDiaFrac_BlkCrs.h"
#include "Mat/BCFlag_Blk.h"
#include "Mat/Vector_Blk.h"

#include "Eqn/ker_emat_tri.h"

using namespace ls;

// APICCG@ŉ֐̐錾
bool Solve_ILU_CG(double& conv_ratio, unsigned int& iteration, std::vector<double>& conv_history,
				const ls::CMatDia_BlkCrs& mat, ls::CVector_Blk& r_vec, ls::CVector_Blk& u_vec,
				const ls::CMatDiaFrac_BlkCrs& frac);


// `ê̗Lvf@UsAꎟɑފ֐
// Rp`1vf
// alpha : vXIy[^̌W
// source : O͍̒萔
bool AddLinearSystem_Poisson(
		double thickness, double lambda, double myu,
		double  rho, double g_x, double g_y,
		ls::CMatDia_BlkCrs& mat, ls::CVector_Blk& residual, 
		const field::CField& val_field,
		const field::CElemArySet& eas, const field::CNodeAry& na )
{
	unsigned int id_ea = val_field.GetID_ElemAry();
	assert( eas.IsID(id_ea) );
	const field::CElemAry& ea = eas.GetEA(id_ea);

	if( ea.ElemType() !=  field::CElemAry::TRI ){
		std::cout << "Not Implimented!!" << std::endl;
		abort();
	}

	const unsigned int nnoel = 3;	// vf̐ߓ_
	const unsigned int ndim = 2;	// Ԃ̎
	
	double g[ndim] = { g_x, g_y };	// d͉x

	unsigned int id_ns_coord = val_field.GetID_NodeSeg_Coord();
	unsigned int id_ns_val   = val_field.GetID_NodeSeg_Value();

	unsigned int noel[nnoel];	// vf̐ߓ_̐ߓ_ԍ

	double emat[nnoel][nnoel][ndim][ndim];	// vfs
	double eforce[nnoel][ndim];		// vfO̓xNg
	double eres[nnoel][ndim];		// vfcxNg

	double coords[nnoel][ndim];		// vfߓ_W
	double edisp[  nnoel][ndim];	// vfߓ_ψ

	double dldx[nnoel][ndim];		// `֐̋Ԕ
	double zero_order_term[nnoel];	// `֐̒萔
	
	unsigned int* tmp_buffer = new unsigned int [mat.NBlkMatRow()];	// }[W̎ɕKvȈꎞz

	for(int ielem=0;ielem<ea.Size();ielem++){

		// vf̐ߓ_ԍĂ
		unsigned int tmp_nnoel;
		ea.GetLpoel(ielem,tmp_nnoel,noel);
		assert( tmp_nnoel == nnoel );

		// ߓ_̍WAlĂ
		for(unsigned int inoel=0;inoel<nnoel;inoel++){
			for(unsigned int idim=0;idim<ndim;idim++){	// ߓ_̍WĂ
				na.GetValueFromNode(noel[inoel],id_ns_coord,idim,coords[inoel][idim]);
			}
			for(unsigned int idim=0;idim<ndim;idim++){	// ߓ_̕ψʂĂ
				na.GetValueFromNode(noel[inoel],id_ns_val  ,idim,  edisp[inoel][idim]);
			}
		}

		////////////////////////////////

		// vfsAcOŏ
		for(unsigned int i=0;i<nnoel*nnoel*ndim*ndim;i++){ *(&emat[0][0][0][0]+i) = 0.0; }
		for(unsigned int i=0;i<           nnoel*ndim;i++){ *(&eforce[0][0]    +i) = 0.0; }

		////////////////////////////////

		// ʐς߂
		double area = TriArea(coords[0],coords[1],coords[2]);

		// `֐̂߂
		TriDlDx(dldx, zero_order_term,   coords[0], coords[1], coords[2]);

		////////////////////////////////

		// s߂
		for(unsigned int inoel=0;inoel<nnoel;inoel++){
		for(unsigned int jnoel=0;jnoel<nnoel;jnoel++){
		   double dtmp1 = 0.0;
		   for(unsigned int idim=0;idim<ndim;idim++){
			  for(unsigned int jdim=0;jdim<ndim;jdim++){
				 emat[inoel][jnoel][idim][jdim] 
					+= area*( lambda*dldx[inoel][idim]*dldx[jnoel][jdim]+myu*dldx[jnoel][idim]*dldx[inoel][jdim] );
			  }
			  dtmp1 += dldx[inoel][idim]*dldx[jnoel][idim];
		   }
		   for(unsigned int idim=0;idim<ndim;idim++){
			  emat[inoel][jnoel][idim][idim] += area*myu*dtmp1;
		   }
		}
		}

		// O̓xNg߂
		for(unsigned int inoel=0;inoel<nnoel;inoel++){
			for(unsigned int idim=0;idim<ndim;idim++){
				eforce[inoel][idim] = area*rho*g[idim]/3.0;
			}
		}

		// vfcxNg߂
		for(unsigned int inoel=0;inoel<nnoel;inoel++){
			for(unsigned int idim=0;idim<ndim;idim++){
				eres[inoel][idim] = eforce[inoel][idim];
			}
			for(unsigned int jnoel=0;jnoel<nnoel;jnoel++){
				for(unsigned int idim=0;idim<ndim;idim++){
				for(unsigned int jdim=0;jdim<ndim;jdim++){
					eres[inoel][idim] -= emat[inoel][jnoel][idim][jdim]*edisp[jnoel][jdim];
				}
				}
			}
		}

		////////////////////////////////
		
		mat.Mearge(nnoel,noel, nnoel,noel, // S̍sɗvfs}[W
			ndim*ndim,&emat[0][0][0][0], 
			tmp_buffer);
		for(unsigned int inoel=0;inoel<nnoel;inoel++){	// vfc}[W
			for(unsigned int idim=0;idim<ndim;idim++){
				residual.AddValue(noel[inoel],idim,eres[inoel][idim]);
			}
		}
	}
	delete[] tmp_buffer;

	return true;
}

void main(){

	field::CNodeAry NA;
	field::CElemArySet EAS;

	ReadMesh("tri_beam.in",NA,EAS);	// bVt@C̓ǂݍ

	field::CField val_field(field::CField::VECTOR,"DISP",  1, NA, EAS);	// ID=1̗vfɃxNgǉ
	field::CField bc0_field(val_field,"BC0_DISP"        ,  8, NA, EAS);	// ID=8̗vfval field̋E̒ǉ

	val_field.SetValueRandom(NA,EAS);		// val_fieldɗݒ
	bc0_field.SetValue(0.0, 0, NA, EAS);	// bc0_field1ݒ
	bc0_field.SetValue(0.0, 1, NA, EAS);	// bc0_field1ݒ

	////////////////////////////////////////////////////////////////

	const unsigned int nnode = NA.Size();
	const unsigned int ndim = 2;

	// E̐ݒ
	CBCFlag bc_flag(nnode,ndim);
	{
		bc_flag.Clear();	// tO
		bc0_field.BoundaryCondition(0,bc_flag,NA,EAS);	// bc0_fieldŌ`Eɐݒ
		bc0_field.BoundaryCondition(1,bc_flag,NA,EAS);	// bc1_fieldŒ苫Eɐݒ
	}

	// Aꎟ̐錾蓖
	CVector_Blk residual(nnode,ndim);
	CVector_Blk update(  nnode,ndim);
	CMatDia_BlkCrs mat(  nnode,ndim);
	{
		unsigned int id_ea = val_field.GetID_ElemAry();
		assert( EAS.IsID(id_ea) );
		mat.AddPattern(EAS.GetEA(id_ea));	// s̔[p^[ݒ
	}
	CMatDiaFrac_BlkCrs frac(0,mat);	// ILUs̏A[p^[̐ݒ
	
	////////////////

	{	// Aꎟւ̒l̐ݒ
		mat.SetZero();
		residual.SetVectorZero();
		const double young = 1000.0;	// O
		const double poisson = 0.3;		// |A\
		const double thickness = 0.1;	// ̌
		const double rho = 1.0;		// x
		const double g_x = 0.0;		// x̏d
		const double g_y = -5.0;	// y̏d
		const double myu = young * 0.5 / ( 1.0+poisson );
		const double lambda0 = young * poisson / ( (1.0+poisson)*(1.0-2.0*poisson) );
		// ʘc̏ꍇ
//		const double lambda = lambda0;
		// ʉ͂̏ꍇ
		const double lambda = lambda0*2.0*myu / ( lambda0 + 2.0*myu );
		// `ê̘Aꎟ𑫂킹
		AddLinearSystem_Poisson(
			thickness, lambda, myu, 
			rho, g_x, g_y, 
			mat,residual,
			val_field,
			EAS,NA);
		mat.SetBoundaryCondition(bc_flag);	// sւ̋E̐ݒ
		bc_flag.SetZeroToBCDof(residual);	// cւ̋E̐ݒ
	}
	frac.SetValue(mat); // ILUOsւ̒l̐ݒILU

	////////////////

	{	// ICCG@ŘAꎟ
		double conv_ratio = 1.0e-6;	// 臒l
		unsigned int iteration = 10000;	// ő唽
		std::vector<double> conv_history;
		Solve_ILU_CG(conv_ratio,iteration, conv_history,
			mat,residual,update,frac);
		std::cout << "Iteration : " << iteration << "   Convarge Ratio : " << conv_ratio << std::endl;
	}

	////////////////

	// l̍XV
	NA.AddValueToNodeSegment(val_field.GetID_NodeSeg_Value(),update,1.0);

	////////////////////////////////////////////////////////////////

	WriteMesh("res.in",NA,EAS,val_field);	// t@Cւ̕ۑ
	system("solid_disp_viewer.exe res.in");	// r[[J

	std::cout << "End Of Analysis" << std::endl;
//	getchar();
}

// ILUOCG@ōs
bool Solve_ILU_CG(double& conv_ratio, unsigned int& iteration, std::vector<double>& conv_history,
				const ls::CMatDia_BlkCrs& mat, ls::CVector_Blk& r_vec, ls::CVector_Blk& u_vec,
				const ls::CMatDiaFrac_BlkCrs& frac){

	const double conv_ratio_tol = conv_ratio;
	const unsigned int mx_iter = iteration;

	const unsigned int nnode = mat.NBlkMatCol();
	const unsigned int nlen = mat.LenBlkCol();

	conv_history.reserve(mx_iter);
	conv_history.resize(0);

	u_vec.SetVectorZero();	
	
	double sq_inv_norm_res0;
	{	// čvZ
		const double sq_norm_res0 = r_vec*r_vec;
		std::cout << "Initial Residual : " << sqrt(sq_norm_res0) << std::endl;
		conv_history.push_back(1.0);
		if( sq_norm_res0 < 1.0e-30 ){	// c\Ζ߂
			conv_ratio = 0.0;
			iteration = 0;
			return true;			
		}
		sq_inv_norm_res0 = 1.0 / sq_norm_res0;
	}

	ls::CVector_Blk p_vec(nnode,nlen);
	ls::CVector_Blk z_vec(nnode,nlen);

	z_vec = r_vec;
	frac.Solve(z_vec);
	p_vec = z_vec;

	double product_rz;
	for(unsigned int iitr=0;iitr<mx_iter;iitr++){

		double alpha;
		{	// alphǎvZ
			product_rz = r_vec * z_vec;
			mat.MatVec(1.0,p_vec,0.0,z_vec);
			double tmp2 = p_vec*z_vec;
			alpha = product_rz / tmp2;
		}

		// ̍XV
		u_vec.AXPY(alpha, p_vec);

		// c̍XV
		r_vec.AXPY(-alpha,z_vec);

		{	// 
			double sq_norm_res = r_vec*r_vec;
			const double ratio = sqrt(sq_norm_res * sq_inv_norm_res0);
//			std::cout << iitr << " " << ratio << std::endl;
			conv_history.push_back(ratio);
			if( sq_norm_res * sq_inv_norm_res0 < conv_ratio_tol*conv_ratio_tol ){
				conv_ratio = sqrt( sq_norm_res * sq_inv_norm_res0 );
				iteration = iitr;
				return true;
			}
		}
	
		double beta;
		{	// betǎvZ
			z_vec = r_vec;
			frac.Solve(z_vec);
			double tmp2 = r_vec*z_vec;
			beta = tmp2/product_rz;
		}

		// TxNgXV
		p_vec *= beta;
		p_vec += z_vec;
	}
	return false;
}
